import warnings
from typing import Dict, List, Literal, Optional, Union

import pandas as pd
from sgqlc.operation import Operation

from ML_management.executor import BaseExecutor
from ML_management.graphql import schema
from ML_management.graphql.schema import (
    ExecutorInfo,
    ExecutorVersionInfo,
    ObjectFilterSettings,
    ObjectVersionFilterSettings,
    TagFilterSettings,
    TimestampInterval,
)
from ML_management.graphql.send_graphql_request import send_graphql_request
from ML_management.mlmanagement.visibility_options import VisibilityOptions
from ML_management.model.model_type_to_methods_map import ModelMethodName
from ML_management.sdk.object import (
    delete_object,
    delete_object_tag,
    delete_object_version,
    delete_object_version_tag,
    get_object_version,
    reset_object_tags,
    reset_object_version_tags,
    set_object_description,
    set_object_tags,
    set_object_version_description,
    set_object_version_tags,
    set_object_version_visibility,
    set_object_visibility,
)
from ML_management.sdk.sdk import _entity, _print_params_by_schema, _to_datetime
from ML_management.types.model_type import ModelType


def get_executor_from_name(name: str) -> ExecutorInfo:
    """
    Get executor.

    Parameters
    ----------
    name: str
        Name of the executor.

    Returns
    -------
    ExecutorInfo
        ExecutorInfo instance with meta information.
    """
    warnings.warn(
        "Function set_dataset_loader_version_visibility is DEPRECATED and will be REMOVED in future releases. "
        "Use set_object_version_visibility",
        DeprecationWarning,
    )
    op = Operation(schema.Query)
    executor_from_name = op.executor_from_name(name=name)
    _entity(executor_from_name)
    executor = send_graphql_request(op=op, json_response=False).executor_from_name
    return executor


def list_executor() -> pd.DataFrame:
    """
    List available executors.

    Returns
    -------
    pd.DataFrame
        Pandas dataframe with list of available executors.
    """
    op = Operation(schema.Query)
    _entity(op.list_executor)
    json_data = send_graphql_request(op)
    df = pd.DataFrame.from_dict(json_data["listExecutor"])
    if not df.empty:
        df = _to_datetime(df, ["creationTimestamp", "lastUpdatedTimestamp"])
    return df


def set_executor_tags(name: str, key: str, values: list[str]) -> ExecutorInfo:
    """
    Set executor tag.

    Parameters
    ----------
    name: str
        Name of the model.
    key: str
        Key tag.
    values: list[str]
        Value tag.

    Returns
    -------
    ExecutorInfo
        Executor instance with meta information.
    """
    warnings.warn(
        "Function set_executor_tags is DEPRECATED and will be REMOVED in future releases. Use set_object_tags",
        DeprecationWarning,
    )

    aggr_id = get_executor_from_name(name).aggr_id
    return set_object_tags(aggr_id=aggr_id, key=key, values=values, model_type=ModelType.EXECUTOR)


def reset_executor_tags(name: str, key: str, values: list[str], new_key: Optional[str] = None) -> ExecutorInfo:
    """
    Reset executor tag.

    Parameters
    ----------
    name: str
        Name of the model.
    key: str
        Key tag.
    values: list[str]
        Value tag.
    new_key: Optional[str] = None
        New key of a tag.


    Returns
    -------
    ExecutorInfo
        Executor instance with meta information.
    """
    warnings.warn(
        "Function reset_executor_tags is DEPRECATED and will be REMOVED in future releases. Use reset_object_tags",
        DeprecationWarning,
    )
    aggr_id = get_executor_from_name(name).aggr_id
    return reset_object_tags(aggr_id=aggr_id, key=key, values=values, new_key=new_key, model_type=ModelType.EXECUTOR)


def delete_executor_tag(name: str, key: str, value: Optional[str] = None) -> ExecutorInfo:
    """
    Delete executor tag.

    Parameters
    ----------
    name: str
        Name of the model.
    key: str
        Key tag.
    value: Optional[str]=None
        value tag.
    Returns
    -------
    ExecutorInfo
        Executor instance with meta information.
    """
    warnings.warn(
        "Function delete_executor_tag is DEPRECATED and will be REMOVED in future releases. Use delete_object_tag",
        DeprecationWarning,
    )

    aggr_id = get_executor_from_name(name).aggr_id
    return delete_object_tag(aggr_id=aggr_id, key=key, value=value, model_type=ModelType.EXECUTOR)


def set_executor_description(name: str, description: str) -> ExecutorInfo:
    """
    Set executor description.

    Parameters
    ----------
    name: str
        Name of the model.
    description: str
        Description model.

    Returns
    -------
    ExecutorInfo
        Executor instance with meta information.
    """
    warnings.warn(
        "Function set_executor_description is DEPRECATED and will be REMOVED in future releases. "
        "Use set_object_description",
        DeprecationWarning,
    )

    aggr_id = get_executor_from_name(name).aggr_id
    return set_object_description(aggr_id=aggr_id, description=description, model_type=ModelType.EXECUTOR)


def set_executor_visibility(
    name: str, visibility: Union[Literal["private", "public"], VisibilityOptions]
) -> ExecutorInfo:
    """
    Set executor visibility.

    Parameters
    ----------
    name: str
        Name of the executor.
    visibility: Union[Literal['private', 'public'], VisibilityOptions]
        Visibility of the executor.

    Returns
    -------
    ExecutorInfo
        Executor instance with meta information.
    """
    warnings.warn(
        "Function set_executor_visibility is DEPRECATED and will be REMOVED in future releases. "
        "Use set_object_visibility",
        DeprecationWarning,
    )

    aggr_id = get_executor_from_name(name).aggr_id
    return set_object_visibility(aggr_id=aggr_id, visibility=visibility, model_type=ModelType.EXECUTOR)


def delete_executor(executor_name: str) -> bool:
    """
    Delete executor and all of it's versions.

    Parameters
    ----------
    executor_name: str
        Name of the executor to delete.

    Returns
    -------
    bool
        Operation success status.
    """
    warnings.warn(
        "Function delete_executor is DEPRECATED and will be REMOVED in future releases. Use delete_object",
        DeprecationWarning,
    )

    aggr_id = get_executor_from_name(executor_name).aggr_id

    return delete_object(aggr_id, ModelType.EXECUTOR)


def list_executor_version(name: str) -> pd.DataFrame:
    """
    List available versions of the executor with such name.

    Parameters
    ----------
    name: str
        Name of the executor.

    Returns
    -------
    pd.DataFrame
        Pandas dataframe with a list of available executor versions.
    """
    warnings.warn(
        "Function set_dataset_loader_visibility is DEPRECATED and will be REMOVED in future releases. "
        "Use set_object_visibility",
        DeprecationWarning,
    )
    op = Operation(schema.Query)
    base_query = op.executor_from_name(name=name).list_executor_version
    base_query.version()
    base_query.creation_timestamp()
    base_query.name()
    json_data = send_graphql_request(op)

    df = pd.DataFrame.from_dict(json_data["executorFromName"]["listExecutorVersion"])
    df = _to_datetime(df, ["creationTimestamp"])

    return df.sort_values(by=["version"], ignore_index=True)


def print_model_schema_for_executor(
    executor_aggr_id: int, models: List[dict], executor_version: Optional[int] = None
) -> None:
    """
    Print model schema for particular executor.

    Parameters
    ----------
    executor_aggr_id: int
        Id of the executor.
    models: List[dict]
        Necessary information about the model.

        Example1::

            [
                {"aggr_id": 1}
            ]

        Example2::

            [
                {
                    "aggr_id": 1, # int
                    "role": "role1", # Optional[str]
                    "version": version # Optional[int]
                },
                {
                    "aggr_id": 2, # int
                    "role": "role2", # Optional[str]
                    "version": version # Optional[int]
                }
            ]
    executor_version: Optional[int] = None
        Version of the executor. Default: None, "latest" version is used.
    """
    models_methods_schemas = _get_model_schema_for_executor(
        executor_aggr_id=executor_aggr_id,
        executor_version=executor_version,
        models=models,
    )
    for model_methods_schemas in models_methods_schemas:
        role = model_methods_schemas["role"]
        for model in models:
            if ("role" in model and model["role"] == role) or "role" not in model:
                aggr_id = model["aggr_id"]

        print(f"Model aggr_id: {aggr_id}, model role: {role}")
        for model_methods_schema in model_methods_schemas["listMethodSchemas"]:
            _print_params_by_schema(
                model_methods_schema["jsonSchema"], ModelMethodName(model_methods_schema["schemaName"]).name
            )
        print()


def cancel_build_job_for_executor_version(aggr_id: int, executor_version: int) -> bool:
    """
    Cancel running or planned build job of executor's image.

    Parameters
    ----------
    aggr_id: int
        Id of the executor.
    executor_version: int
        The version of the executor.

    Returns
    -------
    bool
        Operation success status.
    """
    op = Operation(schema.Mutation)
    op.cancel_build_job_for_executor_version(aggr_id=aggr_id, version=executor_version)
    return send_graphql_request(op)["cancelBuildJobForExecutorVersion"]


def delete_executor_version(executor_name: str, executor_version: int):
    """
    Delete version of a executor.

    Parameters
    ----------
    executor_name: str
        The name of the executor.
    executor_version: int
        The version of the executor.

    Returns
    -------
    None
    """
    warnings.warn(
        "Function set_dataset_loader_version_visibility is DEPRECATED and will be REMOVED in future releases. "
        "Use set_object_version_visibility",
        DeprecationWarning,
    )
    aggr_id = get_executor_from_name(executor_name).aggr_id

    return delete_object_version(aggr_id, executor_version, ModelType.EXECUTOR)


def _get_model_schema_for_executor(executor_aggr_id: int, executor_version: Optional[int], models: list) -> dict:
    role_models = []

    for model in models:
        aggr_id = model["aggr_id"]
        model_version = model.get("version", None)
        if model_version is None:
            model_version = get_object_version(aggr_id).version
        model_role = model.get("role", BaseExecutor.DEFAULT_ROLE)
        current_model = schema.ObjectIdVersionInput(aggr_id=aggr_id, version=model_version)
        current_role = schema.RoleObjectVersionInput(role=model_role, object_version=current_model)

        role_models.append(current_role)

    op = Operation(schema.Query)

    _executor_version = schema.ObjectIdVersionOptionalInput(aggr_id=executor_aggr_id, version=executor_version)
    base_query = (
        op.executor_version_from_aggr_id_version(executor_version=_executor_version)
        .job_json_schema_for_models(models=role_models)
        .list_role_model_method_schemas
    )

    base_query.role()
    base_query.list_method_schemas.schema_name()
    base_query.list_method_schemas.json_schema()

    json_data = send_graphql_request(op)
    return json_data["executorVersionFromAggrIdVersion"]["jobJsonSchemaForModels"]["listRoleModelMethodSchemas"]


def print_executor_schema(aggr_id: int, version: Optional[int] = None) -> None:
    """
    Print executor schema.

    Parameters
    ----------
    aggr_id: int
        Id of the executor.
    version: Optional[int] = None
        Version of the executor. Default: None, "latest" version is used.
    """
    op = Operation(schema.Query)
    executor_version = schema.ObjectIdVersionOptionalInput(aggr_id=aggr_id, version=version)
    base_query = op.executor_version_from_aggr_id_version(executor_version=executor_version)
    base_query.executor_method_schema()
    json_data = send_graphql_request(op)

    json_data = json_data["executorVersionFromAggrIdVersion"]["executorMethodSchema"]
    _print_params_by_schema(json_schema=json_data, schema_type="Executor")


def print_executor_roles(aggr_id: int, version: Optional[int] = None) -> None:
    """
    Print the roles required by the executor.

    Parameters
    ----------
    aggr_id: int
        Id of the executor.
    version: Optional[int] = None
        Version of the executor. Default: None, "latest" version is used.
    """
    op = Operation(schema.Query)
    executor_version = schema.ObjectIdVersionOptionalInput(aggr_id=aggr_id, version=version)
    base_query = op.executor_version_from_aggr_id_version(executor_version=executor_version)
    base_query.desired_model_methods()
    base_query.desired_dataset_loader_methods()
    json_data = send_graphql_request(op)
    print("Desired model methods:", json_data["executorVersionFromAggrIdVersion"]["desiredModelMethods"])
    print(
        "Desired dataset loader methods:", json_data["executorVersionFromAggrIdVersion"]["desiredDatasetLoaderMethods"]
    )


def get_required_model_classes_by_executor(
    aggr_id: int, version: Optional[int] = None
) -> Dict[str, List[Dict[str, str]]]:
    """
    Return the names of classes for the model to be inherited from, by the name of the executor.

    Parameters
    ----------
    aggr_id: int
        Id of the executor.
    version: Optional[int] = None
        Version of the executor. Default: None, "latest" version is used.

    Returns
    -------
    List[str]:
        List of model class names to be inherited from.
    """
    op = Operation(schema.Query)
    executor_version = schema.ObjectIdVersionOptionalInput(aggr_id=aggr_id, version=version)
    base_query = op.executor_version_from_aggr_id_version(executor_version=executor_version)
    base_query.desired_model_patterns()
    json_data = send_graphql_request(op)
    return json_data["executorVersionFromAggrIdVersion"]["desiredModelPatterns"]


def get_required_dataset_loader_classes_by_executor(
    aggr_id: int, version: Optional[int] = None
) -> Dict[str, List[Dict[str, str]]]:
    """
    Return the names of classes for the dataset loader to be inherited from, by the name of the executor.

    Parameters
    ----------
    aggr_id: int
        Id of the executor.
    version: Optional[int] = None
        Version of the executor. Default: None, "latest" version is used.

    Returns
    -------
    List[str]:
        List of dataset loader class names to be inherited from.
    """
    op = Operation(schema.Query)
    executor_version = schema.ObjectIdVersionOptionalInput(aggr_id=aggr_id, version=version)
    base_query = op.executor_version_from_aggr_id_version(executor_version=executor_version)
    base_query.desired_dataset_loader_patterns()
    json_data = send_graphql_request(op)
    return json_data["executorVersionFromAggrIdVersion"]["desiredDatasetLoaderPatterns"]


def set_executor_version_description(name: str, version: int, description: str) -> ExecutorVersionInfo:
    """
    Set executor version description.

    Parameters
    ----------
    name: str
        Name of the executor.
    version: int
        Version of the executor.
    description: str
        Description executor version.

    Returns
    -------
    ExecutorVersionInfo
        Executor version instance with meta information.
    """
    warnings.warn(
        "Function set_executor_version_description is DEPRECATED and will be REMOVED in future releases. "
        "Use set_object_version_description",
        DeprecationWarning,
    )

    aggr_id = get_executor_from_name(name).aggr_id
    return set_object_version_description(
        aggr_id=aggr_id, version=version, description=description, model_type=ModelType.EXECUTOR
    )


def set_executor_version_visibility(
    name: str,
    version: int,
    visibility: Union[Literal["private", "public"], VisibilityOptions],
) -> ExecutorVersionInfo:
    """
    Set executor version visibility.

    Parameters
    ----------
    name: str
        Name of the executor.
    version: int
        Version of the executor.
    visibility: Union[Literal['private', 'public'], VisibilityOptions]
        Visibility executor version.

    Returns
    -------
    ExecutorVersionInfo
        Executor version instance with meta information.
    """
    warnings.warn(
        "Function set_executor_version_visibility is DEPRECATED and will be REMOVED in future releases. "
        "Use set_object_version_visibility",
        DeprecationWarning,
    )

    aggr_id = get_executor_from_name(name).aggr_id
    return set_object_version_visibility(
        aggr_id=aggr_id, version=version, visibility=visibility, model_type=ModelType.EXECUTOR
    )


def set_executor_version_tags(name: str, version: int, key: str, values: list[str]) -> ExecutorVersionInfo:
    """
    Set executor version tags.

    Parameters
    ----------
    name: str
        Name of the executor.
    version: int
        Version of the executor.
    key: str
        Key tag.
    values: list[str]
        Value tag.

    Returns
    -------
    ExecutorVersionInfo
        Executor version instance with meta information.
    """
    warnings.warn(
        "Function set_executor_version_tags is DEPRECATED and will be REMOVED in future releases. "
        "Use set_object_version_tags",
        DeprecationWarning,
    )

    aggr_id = get_executor_from_name(name).aggr_id
    return set_object_version_tags(
        aggr_id=aggr_id, version=version, key=key, values=values, model_type=ModelType.EXECUTOR
    )


def reset_executor_version_tags(
    name: str, version: int, key: str, values: list[str], new_key: Optional[str] = None
) -> ExecutorVersionInfo:
    """
    Reset executor version tags.

    Parameters
    ----------
    name: str
        Name of the executor.
    version: int
        Version of the executor.
    key: str
        Key tag.
    values: list[str]
        Value tag.
    new_key: Optional[str] = None
        New key of a tag.

    Returns
    -------
    ExecutorVersionInfo
        Executor version instance with meta information.
    """
    warnings.warn(
        "Function reset_executor_version_tags is DEPRECATED and will be REMOVED in future releases. "
        "Use reset_object_version_tags",
        DeprecationWarning,
    )

    aggr_id = get_executor_from_name(name).aggr_id
    return reset_object_version_tags(
        aggr_id=aggr_id, version=version, key=key, values=values, new_key=new_key, model_type=ModelType.EXECUTOR
    )


def delete_executor_version_tag(name: str, version: int, key: str, value: Optional[str] = None) -> ExecutorVersionInfo:
    """
    Delete executor version tag.

    Parameters
    ----------
    name: str
        Name of the executor.
    version: int
        Version of the executor.
    key: str
        Key tag.
    value: Optional[str]=None
        value tag.

    Returns
    -------
    ExecutorVersionInfo
        Executor version instance with meta information.
    """
    warnings.warn(
        "Function delete_executor_version_tag is DEPRECATED and will be REMOVED in future releases. "
        "Use delete_object_version_tag",
        DeprecationWarning,
    )

    aggr_id = get_executor_from_name(name).aggr_id
    return delete_object_version_tag(
        aggr_id=aggr_id, version=version, key=key, value=value, model_type=ModelType.EXECUTOR
    )


def get_executor_version(name: str, version: Optional[int] = None) -> ExecutorVersionInfo:
    """
    Meta information about the executor version by the executor name and version.

    Parameters
    ----------
    name: str
        Name of the executor.
    version: Optional[int] = None
        Version of the executor. Default: None, "latest" version is used.

    Returns
    -------
    ExecutorVersionInfo
        ExecutorVersion instance with meta information.
    """
    warnings.warn(
        "Function set_dataset_loader_version_visibility is DEPRECATED and will be REMOVED in future releases. "
        "Use set_object_version_visibility",
        DeprecationWarning,
    )
    op = Operation(schema.Query)
    executor_version_choice = schema.ObjectVersionOptionalInput(name=name, version=version)
    base_query = op.executor_version_from_name_version(executor_version=executor_version_choice)
    base_query.name()
    base_query.version()
    base_query.tags()
    base_query.description()
    base_query.creation_timestamp()
    data = send_graphql_request(op, json_response=False).executor_version_from_name_version
    return data


def get_executor_version_conda_env(name: str, version: int) -> dict:
    """
    Condas configuration for the executor version by the executor name and version.

    Parameters
    ----------
    name: str
        Name of the executor.
    version: Optional[int] = None
        Version of the executor. Default: None, "latest" version is used.

    Returns
    -------
    Dict
        Dict with conda configuration.
    """
    warnings.warn(
        "Function set_dataset_loader_version_visibility is DEPRECATED and will be REMOVED in future releases. "
        "Use set_object_version_visibility",
        DeprecationWarning,
    )
    op = Operation(schema.Query)
    _model_version = schema.ObjectVersionOptionalInput(name=name, version=version)
    base_query = op.executor_version_from_name_version(executor_version=_model_version)
    base_query.get_conda_env()
    model_version = send_graphql_request(op, json_response=False)
    return model_version.executor_version_from_name_version.get_conda_env


def get_executor_version_requirements(name: str, version: int) -> list:
    """
    Requirements for the executor version by the executor name and version.

    Parameters
    ----------
    name: str
        Name of the executor.
    version: Optional[int] = None
        Version of the executor. Default: None, "latest" version is used.

    Returns
    -------
    List
        List of requirements.
    """
    warnings.warn(
        "Function set_dataset_loader_version_visibility is DEPRECATED and will be REMOVED in future releases. "
        "Use set_object_version_visibility",
        DeprecationWarning,
    )
    op = Operation(schema.Query)
    _model_version = schema.ObjectVersionOptionalInput(name=name, version=version)
    base_query = op.executor_version_from_name_version(executor_version=_model_version)
    base_query.list_requirements()
    model_version = send_graphql_request(op, json_response=False)
    return model_version.executor_version_from_name_version.list_requirements


def get_latest_executor_version(name: str) -> ExecutorVersionInfo:
    """
    Latest executor version by the executor name.

    Parameters
    ----------
    name: str
        Name of the executor.

    Returns
    -------
    ExecutorVersionInfo
        ExecutorVersion instance with meta information.
    """
    warnings.warn(
        "Function set_dataset_loader_version_visibility is DEPRECATED and will be REMOVED in future releases. "
        "Use set_object_version_visibility",
        DeprecationWarning,
    )
    return get_executor_version(name)


def get_initial_executor_version(name: str) -> ExecutorVersionInfo:
    """
    Initial executor version by the executor name.

    Parameters
    ----------
    name: str
        Name of the executor.

    Returns
    -------
    ExecutorVersionInfo
        ExecutorVersion instance with meta information.
    """
    warnings.warn(
        "Function set_dataset_loader_version_visibility is DEPRECATED and will be REMOVED in future releases. "
        "Use set_object_version_visibility",
        DeprecationWarning,
    )
    op = Operation(schema.Query)
    version = op.executor_from_name(name=name).init_executor_version()
    version.name()
    version.version()
    version.tags()
    version.description()
    executor_version = send_graphql_request(op, json_response=False)
    return executor_version.executor_from_name.init_executor_version


def pagination_executor(
    name: Optional[str] = None,
    tag_key: Optional[str] = None,
    tag_value: Optional[str] = None,
    description: Optional[str] = None,
    visibility: Optional[Union[Literal["private", "public"], VisibilityOptions]] = None,
    owner_ids: Optional[list[str]] = None,
    creation_from: Optional[int] = None,
    creation_to: Optional[int] = None,
    last_updated_from: Optional[int] = None,
    last_updated_to: Optional[int] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> List[ExecutorInfo]:
    """
    Search executors.

    Parameters
    ----------
    name: Optional[str]=None
        Name of the executor.
    tag_key: Optional[str]=None
        Key of the executor tag.
    tag_value: Optional[str]=None
        Value of the executor tag.
    description: Optional[str]=None
        Description of the executor.
    visibility: Optional[Union[Literal['private', 'public'], VisibilityOptions]]=None
        Visibility of executor.
    owner_ids: Optional[list[str]]=None
        Ids of the executor owner.
    creation_from: Optional[int]=None
        Creation timestamp from of the executor.
    creation_to: Optional[int]=None
        Creation timestamp from of the executor.
    last_updated_from: Optional[int]=None
        Last updated timestamp from of the executor.
    last_updated_to: Optional[int]=None
        Last updated timestamp from of the executor.
    limit: Optional[int] = None
        The maximum number of records that will be returned as a result.
    offset: Optional[int] = None
        The number of records that will be skipped before starting the selection.

    Returns
    -------
    List[ExecutorInfo]
        List of ExecutorInfo instance with meta information.
    """
    op = Operation(schema.Query)
    visibility = VisibilityOptions(visibility) if visibility else visibility
    base_query = op.pagination_executor(
        limit=limit,
        offset=offset,
        filter_settings=ObjectFilterSettings(
            name=name,
            description=description,
            visibility=visibility,
            owner_ids=owner_ids,
            tag=TagFilterSettings(key=tag_key, value=tag_value),
            creation_interval=TimestampInterval(start=creation_from, end=creation_to),
            last_updated_interval=TimestampInterval(start=last_updated_from, end=last_updated_to),
        ),
    ).list_executor
    _entity(base_query)
    return send_graphql_request(op, json_response=False).pagination_executor.list_executor


def pagination_executor_version(
    name: str,
    version: Optional[int] = None,
    tag_key: Optional[str] = None,
    tag_value: Optional[str] = None,
    description: Optional[str] = None,
    visibility: Optional[VisibilityOptions] = None,
    owner_ids: Optional[list[str]] = None,
    creation_from: Optional[int] = None,
    creation_to: Optional[int] = None,
    last_updated_from: Optional[int] = None,
    last_updated_to: Optional[int] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> List[ExecutorVersionInfo]:
    """
    Search executor versions.

    Parameters
    ----------
    name: Optional[str]=None
        Name of the executor.
    version: Optional[int] = None
        Version of the executor.
    tag_key: Optional[str]=None
        Key of the executor tag.
    tag_value: Optional[str]=None
        Value of the executor tag.
    description: Optional[str]=None
        Description of the executor version.
    visibility: Optional[str]=None
        Visibility of executor version.
    owner_ids: Optional[list[str]]=None
        Ids of the executor version owner.
    creation_from: Optional[int]=None
        Creation timestamp from of the executor version.
    creation_to: Optional[int]=None
        Creation timestamp from of the executor version.
    last_updated_from: Optional[int]=None
        Last updated timestamp from of the executor version.
    last_updated_to: Optional[int]=None
        Last updated timestamp from of the executor version.
    limit: Optional[int] = None
        The maximum number of records that will be returned as a result.
    offset: Optional[int] = None
        The number of records that will be skipped before starting the selection.

    Returns
    -------
    List[ExecutorVersionInfo]
        List of ExecutorVersionInfo instance with meta information.
    """
    warnings.warn(
        "Function pagination_executor_version is DEPRECATED and will be REMOVED in future releases. "
        "Use pagination_executor_version_from_id",
        DeprecationWarning,
    )
    op = Operation(schema.Query)
    base_query = (
        op.executor_from_name(name=name)
        .pagination_executor_version(
            limit=limit,
            offset=offset,
            filter_settings=ObjectVersionFilterSettings(
                version=version,
                description=description,
                visibility=visibility,
                owner_ids=owner_ids,
                tag=TagFilterSettings(key=tag_key, value=tag_value),
                creation_interval=TimestampInterval(start=creation_from, end=creation_to),
                last_updated_interval=TimestampInterval(start=last_updated_from, end=last_updated_to),
            ),
        )
        .list_executor_version
    )
    base_query.name()
    base_query.version()
    base_query.tags()
    base_query.description()
    return send_graphql_request(
        op, json_response=False
    ).executor_from_name.pagination_executor_version.list_executor_version


def pagination_executor_version_from_id(
    aggr_id: int,
    version: Optional[int] = None,
    tag_key: Optional[str] = None,
    tag_value: Optional[str] = None,
    description: Optional[str] = None,
    visibility: Optional[VisibilityOptions] = None,
    owner_ids: Optional[list[str]] = None,
    creation_from: Optional[int] = None,
    creation_to: Optional[int] = None,
    last_updated_from: Optional[int] = None,
    last_updated_to: Optional[int] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> List[ExecutorVersionInfo]:
    """
    Search executor versions.

    Parameters
    ----------
    aggr_id: int
        Id of the object.
    version: Optional[int] = None
        Version of the executor.
    tag_key: Optional[str]=None
        Key of the executor tag.
    tag_value: Optional[str]=None
        Value of the executor tag.
    description: Optional[str]=None
        Description of the executor version.
    visibility: Optional[str]=None
        Visibility of executor version.
    owner_ids: Optional[list[str]]=None
        Ids of the executor version owner.
    creation_from: Optional[int]=None
        Creation timestamp from of the executor version.
    creation_to: Optional[int]=None
        Creation timestamp from of the executor version.
    last_updated_from: Optional[int]=None
        Last updated timestamp from of the executor version.
    last_updated_to: Optional[int]=None
        Last updated timestamp from of the executor version.
    limit: Optional[int] = None
        The maximum number of records that will be returned as a result.
    offset: Optional[int] = None
        The number of records that will be skipped before starting the selection.

    Returns
    -------
    List[ExecutorVersionInfo]
        List of ExecutorVersionInfo instance with meta information.
    """
    op = Operation(schema.Query)
    base_query = (
        op.executor_from_id(aggr_id=aggr_id)
        .pagination_executor_version(
            limit=limit,
            offset=offset,
            filter_settings=ObjectVersionFilterSettings(
                version=version,
                description=description,
                visibility=visibility,
                owner_ids=owner_ids,
                tag=TagFilterSettings(key=tag_key, value=tag_value),
                creation_interval=TimestampInterval(start=creation_from, end=creation_to),
                last_updated_interval=TimestampInterval(start=last_updated_from, end=last_updated_to),
            ),
        )
        .list_executor_version
    )
    base_query.name()
    base_query.version()
    base_query.tags()
    base_query.description()
    return send_graphql_request(
        op, json_response=False
    ).executor_from_id.pagination_executor_version.list_executor_version



"""Script for upload default dummy datasetloader."""
import os

from ML_management import mlmanagement
from ML_management.mlmanagement.visibility_options import VisibilityOptions


def upload_dummmy_data():
    mlmanagement.log_dataset_loader_src(
        description="Dummy dataset loader",
        registered_name="DummyData",
        model_path=os.path.join(os.path.dirname(os.path.realpath(__file__)), "dummy_dataset_loader"),
        visibility=VisibilityOptions.PUBLIC,
    )



"""Executor template for custom executor."""

from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, Union

from ML_management.dataset_loader import DatasetLoaderPattern
from ML_management.dataset_loader.dataset_loader_pattern_to_methods_map import DatasetLoaderMethodName
from ML_management.executor.patterns import (
    DEFAULT_ROLE,
    ArbitraryDatasetLoaderPattern,
    ArbitraryModelsPattern,
    OneDatasetLoaderPattern,
    OneModelPattern,
)
from ML_management.model.model_type_to_methods_map import ModelMethodName
from ML_management.model.patterns.model_pattern import Model

default_dataset_loader_pattern = OneDatasetLoaderPattern()


class BaseExecutor(ABC):
    """Define custom job executor."""

    DEFAULT_ROLE = DEFAULT_ROLE

    def __init__(
        self,
        executor_models_pattern: Union[OneModelPattern, ArbitraryModelsPattern],
        executor_dataset_loaders_pattern: Union[OneDatasetLoaderPattern, ArbitraryDatasetLoaderPattern, None] = None,
    ) -> None:
        """
        Init Executor class.

        Parameters
        ----------
        executor_models_pattern: Union[OneModelPattern, ArbitraryModelsPattern]
            Pattern of interaction with models.
        executor_datasets_pattern: Union[OneDatasetLoaderPattern, ArbitraryDatasetLoaderPattern
            Pattern of datasets usage. Default: OneDatasetLoaderPattern()

        Returns
        -------
        None

        """
        if executor_dataset_loaders_pattern is None:
            executor_dataset_loaders_pattern = default_dataset_loader_pattern

        if not isinstance(executor_models_pattern, OneModelPattern) and not isinstance(
            executor_models_pattern, ArbitraryModelsPattern
        ):
            raise TypeError(
                "param executor_models_pattern must be instance of OneModelPattern or ArbitraryModelsPattern"
            )
        if not isinstance(executor_dataset_loaders_pattern, OneDatasetLoaderPattern) and not isinstance(
            executor_dataset_loaders_pattern, ArbitraryDatasetLoaderPattern
        ):
            raise TypeError(
                "param executor_datasets_pattern must be instance of OneDatasetPattern or ArbitraryDatasetPattern"
            )
        self.desired_model_methods = executor_models_pattern.serialize()
        self.desired_dataset_loader_methods = executor_dataset_loaders_pattern.serialize()
        for methods_list in self.desired_dataset_loader_methods.values():
            if DatasetLoaderMethodName.get_dataset not in methods_list:
                raise RuntimeError(
                    "Every dataset loader must have DatasetLoaderMethodName.get_dataset method in desired methods."
                )
        # because each class always have init and we do not want init in desired methods written by user.
        for role, methods_list in self.desired_model_methods.items():
            if ModelMethodName.init not in methods_list:
                self.desired_model_methods[role].append(ModelMethodName.init)

        for role, methods_list in self.desired_dataset_loader_methods.items():
            if DatasetLoaderMethodName.init not in methods_list:
                self.desired_dataset_loader_methods[role].append(ModelMethodName.init)

        # That parameters will be set automatically while loading the model.
        self.artifacts: str

        # That parameter will be set automatically in job before the 'execute' func would be executed.
        self.role_dataset_loader_map: Dict[str, DatasetLoaderPattern] = {}
        self.device = None
        self.role_model_map: Dict[str, Model] = {}
        self.role_dataset_map: Dict[str, Any] = {}
        self.dataset_loader_method_parameters_dict: Dict[str, Dict[DatasetLoaderMethodName, dict]] = {}
        self.model_method_parameters_dict: Dict[str, Dict[ModelMethodName, dict]] = {}

    @abstractmethod
    def execute(self, **executor_params):
        """
        Do execution step.

        Attributes
        ----------
        artifacts: str
            local path to artifacts.
        role_model_map: Dict[str, Model]
            role to model map. Default: {}.
        role_dataset_loader_map: Dict[str, DatasetLoaderPattern]
            role to dataset loader map. Default: {}.
        role_dataset_map: Dict[str, Any]
            role to dataset map.
            Dataset is an object obtained as a result of calling the 'get_dataset' function
            from a dataset loader with the same role. Default: {}.
        device: str
            name of the device ('cpu' - CPU, 'cuda' - GPU instance). Default: None.
        model_method_parameters_dict: Dict[str, Dict[ModelMethodName, dict]]
            the dict of parameters for each desired_model_methods.
            One could use it in execute() function like that::

                def execute(self):
                    self.role_model_map["role_model"].train_function(
                        **self.model_method_parameters_dict["role_model"][ModelMethodName.train_function]
                    )

            In that case method 'execute' calls train_function method of the model with corresponding parameters
            for that method. Default: {}.
        dataset_loader_method_parameters_dict: Dict[str, Dict[DatasetLoaderMethodName, dict]]
            the dict of params for each desired_dataset_loader_methods.
            One could use it in execute() function like that::

                def execute(self):
                    self.role_dataset_loader_map["role_dataset"].get_dataset(
                        **self.dataset_loader_method_parameters_dict["role_dataset"][DatasetLoaderMethodName.get_dataset]
                    )

            In that case method 'execute' calls get_dataset method of the dataset loader with corresponding parameters
            for that method. Default: {}.

        Returns
        -------
        None
        """
        raise NotImplementedError

    @property
    def model(self) -> Model:
        """Property returning a single model.

        :return: python model
        """
        if len(self.role_model_map) != 1:
            raise RuntimeError("This attribute can be used only with 1 model")
        return list(self.role_model_map.values())[0]

    @model.setter
    def model(self, model):
        """Property to change the model.

        :param: python model
        """
        if len(self.role_model_map) != 1:
            raise RuntimeError("This attribute can be used only with 1 model")
        self.role_model_map[list(self.role_model_map.keys())[0]] = model

    @property
    def dataset_loader(self) -> DatasetLoaderPattern:
        """Property returning a single dataset loader.

        :return: dataset loader object
        """
        if len(self.role_dataset_loader_map) != 1:
            raise RuntimeError("This attribute can be used only with 1 dataset loader")
        return list(self.role_dataset_loader_map.values())[0]

    @dataset_loader.setter
    def dataset_loader(self, dataset_loader: DatasetLoaderPattern):
        """Property to change the dataset loader.

        :param: dataset loader object
        """
        if len(self.role_dataset_loader_map) != 1:
            raise RuntimeError("This attribute can be used only with 1 dataset loader")
        self.role_dataset_loader_map[list(self.role_dataset_loader_map.keys())[0]] = dataset_loader

    @property
    def model_methods_parameters(self) -> Dict[ModelMethodName, dict]:
        """Property for the dictionary wrapper.

        :return: the dict of parameters for each desired_model_methods
        """
        if len(self.model_method_parameters_dict) != 1:
            raise RuntimeError("This attribute can be used only with 1 model")
        return list(self.model_method_parameters_dict.values())[0]

    @property
    def dataset_loader_methods_parameters(self) -> Dict[DatasetLoaderMethodName, dict]:
        """Property for the dictionary wrapper.

        :return: the dict of parameters for each desired_dataset_loader_methods
        """
        if len(self.dataset_loader_method_parameters_dict) != 1:
            raise RuntimeError("This attribute can be used only with 1 model")
        return list(self.dataset_loader_method_parameters_dict.values())[0]

    @property
    def dataset(self):
        """Property returning a single dataset.

        :return: dataset object
        """
        if len(self.role_dataset_map) != 1:
            raise RuntimeError("This attribute can be used only with 1 dataset")
        return list(self.role_dataset_map.values())[0]

    @dataset.setter
    def dataset(self, dataset: Any):
        """Property to change the dataset.

        :param: dataset object
        """
        if len(self.role_dataset_map) != 1:
            raise RuntimeError("This attribute can be used only with 1 dataset")
        self.role_dataset_map[list(self.role_dataset_map.keys())[0]] = dataset


class BaseExecutorDistributed(BaseExecutor):
    """Define custom job executor for job that can be executed in distributed mode."""

    def __init__(
        self,
        executor_models_pattern: Union[OneModelPattern, ArbitraryModelsPattern],
        executor_dataset_loaders_pattern: Union[OneDatasetLoaderPattern, ArbitraryDatasetLoaderPattern, None] = None,
    ) -> None:
        super().__init__(
            executor_models_pattern=executor_models_pattern,
            executor_dataset_loaders_pattern=executor_dataset_loaders_pattern,
        )
        # attributes related to distributed job execution
        self.nodes: list[str] = []
        self.master_addr: Optional[str] = None
        self.master_port: Optional[int] = None
        self.node_rank: Optional[int] = None

    @property
    def nodes_number(self) -> int:
        return len(self.nodes)



"""Upload model mode."""
from enum import Enum


class UploadModelMode(str, Enum):
    """DEPRECATED.

    DO NOT USE. EXISTS FOR BACKWARD COMPATIBILITY ONLY.
    WILL BE DELETED IN FUTURE RELEASES.
    """

    none = "none"
    new_model = "new_model"  # log model as a new one with new name
    new_version = "new_version"  # log model as a new version of existing model



from ML_management.jsonschema_inference import jsonschema_exceptions, jsonschema_inference

SkipJsonSchema = jsonschema_inference.SkipJsonSchema
infer_jsonschema = jsonschema_inference.infer_jsonschema
DictKeysMustBeStringsError = jsonschema_exceptions.DictKeysMustBeStringsError
FunctionContainsVarArgsError = jsonschema_exceptions.FunctionContainsVarArgsError
FunctionContainsVarKwArgsError = jsonschema_exceptions.FunctionContainsVarKwArgsError
InvalidStructureAnnotationError = jsonschema_exceptions.InvalidStructureAnnotationError
NoAnnotationError = jsonschema_exceptions.NoAnnotationError
UnsupportedTypeError = jsonschema_exceptions.UnsupportedTypeError



__all__ = ["COLLECTORS", "REMOTE_COLLECTORS"]

from ML_management.collectors.collectors import DATA_COLLECTORS, DATA_REMOTE_COLLECTORS

COLLECTORS = DATA_COLLECTORS

REMOTE_COLLECTORS = DATA_REMOTE_COLLECTORS



import locale
import posixpath
from typing import Any, Optional

import httpx
from sgqlc.operation import Operation

from ML_management.session import AuthSession
from ML_management.variables import get_server_url


def send_graphql_request(op: Operation, json_response: bool = True, timeout: Optional[int] = 60) -> Any:
    """Send request to server and process the response."""
    json_data = AuthSession().sgqlc_request(op, timeout)

    if "errors" in json_data:
        server_url = get_server_url()
        try:
            if locale.getdefaultlocale()[0][:2] == "ru":
                url_base = posixpath.join(server_url, "locales/ru/ru.json")
            else:
                url_base = posixpath.join(server_url, "locales/en/en.json")
        except Exception:
            # if there is no locale file use english by default.
            url_base = posixpath.join(server_url, "locales/en/en.json")
        translation = httpx.get(url_base).json()

        error_message = json_data["errors"][0]["message"].split(",")[0]

        try:
            message_type, message_value = error_message.split(".")
        except Exception:
            raise Exception(error_message) from None

        if message_type not in translation:
            raise Exception(message_type)

        formatted_translated_message = translation[message_type][message_value]
        if (
            ("extensions" in json_data["errors"][0])
            and ("params" in json_data["errors"][0]["extensions"][error_message])
            and (json_data["errors"][0]["extensions"][error_message]["params"] is not None)
        ):
            raise Exception(
                formatted_translated_message.format().format(
                    **json_data["errors"][0]["extensions"][error_message]["params"]
                )
            )
        raise Exception(formatted_translated_message)

    if json_response:
        return json_data["data"]
    else:
        return op + json_data



class MLMBaseError(Exception):
    """Base exception for all custom exceptions."""

    pass


class MLMServerError(MLMBaseError):
    """Base exception for all server mlmanager exceptions."""

    pass


class RegistryError(MLMBaseError):
    """Base exception for all registry exceptions."""

    pass


class MLMClientError(MLMBaseError):
    """Base exception for all client-side specific exceptions."""

    pass



import warnings
from dataclasses import dataclass
from typing import Union

from ML_management.dataset_loader import DatasetLoaderPattern
from ML_management.executor import BaseExecutor
from ML_management.mlmanagement.model_type import ModelType
from ML_management.model import Model

warnings.warn(
    "ObjectMetaInfo and LoadedObject will be moved to ML_management.types.metainfo in future versions.",
    DeprecationWarning,
)


@dataclass
class ObjectMetaInfo:
    """
    Represents a logged object with its metadata.

    Returned when logging an entity.

    Attributes:
        name (str): Name of the entity.
        version (int): Version of the entity.
        hash_artifacts (str): Hash of the entity artifacts.
        model_type (ModelType): Entity type.
    """

    name: str
    aggr_id: int
    version: int
    hash_artifacts: str
    model_type: ModelType


@dataclass
class LoadedObject:
    """
    Represents a loaded object with its metadata.

    Returned when loading an entity.

    Attributes:
        local_path (str): Local path where the object is stored.
        loaded_object (Union[BaseExecutor, DatasetLoaderPattern, Model]): The loaded class instance.
        metainfo (ObjectMetaInfo): Metadata information about the loaded object.
    """

    local_path: str
    loaded_object: Union[BaseExecutor, DatasetLoaderPattern, Model]
    metainfo: ObjectMetaInfo



from typing import Literal, Optional, Union

import pandas as pd
from sgqlc.operation import Operation

from ML_management.graphql.schema import (
    DatasetLoaderInfo,
    DatasetLoaderVersionInfo,
    ExecutorInfo,
    ExecutorVersionInfo,
    ModelInfo,
    ModelVersionInfo,
    UpdateObjectForm,
    UpdateObjectVersionForm,
    schema,
)
from ML_management.graphql.send_graphql_request import send_graphql_request
from ML_management.mlmanagement.model_type import ModelType
from ML_management.mlmanagement.visibility_options import VisibilityOptions
from ML_management.sdk.sdk import _entity, _to_datetime

_object_map = {
    ModelType.MODEL: ModelInfo,
    ModelType.DATASET_LOADER: DatasetLoaderInfo,
    ModelType.EXECUTOR: ExecutorInfo,
}

_object_version_map = {
    ModelType.MODEL: ModelVersionInfo,
    ModelType.DATASET_LOADER: DatasetLoaderVersionInfo,
    ModelType.EXECUTOR: ExecutorVersionInfo,
}


def _base_get_version(op, aggr_id, version, model_type):
    _version = schema.ObjectIdVersionOptionalInput(aggr_id=aggr_id, version=version)

    if model_type == ModelType.MODEL:
        base_query = op.model_version_from_aggr_id_version(model_version=_version)

    elif model_type == ModelType.EXECUTOR:
        base_query = op.executor_version_from_aggr_id_version(executor_version=_version)

    else:
        base_query = op.dataset_loader_version_from_aggr_id_version(dataset_loader_version=_version)

    return base_query


def get_object(
    aggr_id: int, model_type: Union[Literal["model", "executor", "dataset_loader"], ModelType] = ModelType.MODEL
) -> ModelInfo | ExecutorInfo | DatasetLoaderInfo:
    """
    Get object.

    Parameters
    ----------
    aggr_id: int
        Id of the object.

    model_type: Union[Literal['model', 'executor', 'dataset_loader'] , ModelType]
        Type of the entity. Possible values: ModelType.MODEL | ModelType.EXECUTOR | ModelType.DATASET_LOADER

    Returns
    -------
    ModelInfo|ExecutorInfo|DatasetLoaderInfo
        Instance with meta information.
    """
    op = Operation(schema.Query)
    if model_type == ModelType.MODEL:
        model_from_id = op.model_from_id(aggr_id=aggr_id)
        _entity(model_from_id)
        model = send_graphql_request(op=op, json_response=False).model_from_id
        return model
    elif model_type == ModelType.EXECUTOR:
        executor_from_name = op.executor_from_id(aggr_id=aggr_id)
        _entity(executor_from_name)
        executor = send_graphql_request(op=op, json_response=False).executor_from_id
        return executor
    else:
        dataset_loader_from_name = op.dataset_loader_from_id(aggr_id=aggr_id)
        _entity(dataset_loader_from_name)
        dataset_loader = send_graphql_request(op=op, json_response=False).dataset_loader_from_id
        return dataset_loader


def list_object_version(
    aggr_id: int, model_type: Union[Literal["model", "executor", "dataset_loader"], ModelType] = ModelType.MODEL
) -> pd.DataFrame:
    """
    List available versions of the object with such aggr_id.

    Parameters
    ----------
    aggr_id: int
        Id of the object.
    model_type: Union[Literal['model', 'executor', 'dataset_loader'] , ModelType]
        Type of the entity. Possible values: ModelType.MODEL | ModelType.EXECUTOR | ModelType.DATASET_LOADER

    Returns
    -------
    pd.DataFrame
        Pandas dataframe with a list of available object versions.
    """
    model_type = ModelType(model_type)

    op = Operation(schema.Query)

    if model_type == ModelType.MODEL:
        base_query = op.model_from_id(aggr_id=aggr_id).list_model_version

    elif model_type == ModelType.EXECUTOR:
        base_query = op.executor_from_id(aggr_id=aggr_id).list_executor_version

    else:
        base_query = op.dataset_loader_from_id(aggr_id=aggr_id).list_dataset_loader_version

    base_query.version()
    base_query.creation_timestamp()
    base_query.name()
    json_data = send_graphql_request(op)

    if model_type == ModelType.MODEL:
        df = pd.DataFrame.from_dict(json_data["modelFromId"]["listModelVersion"])

    elif model_type == ModelType.EXECUTOR:
        df = pd.DataFrame.from_dict(json_data["modelFromId"]["listExecutorVersion"])

    else:
        df = pd.DataFrame.from_dict(json_data["modelFromId"]["listDatasetLoaderVersion"])

    df = _to_datetime(df, ["creationTimestamp"])

    return df.sort_values(by=["version"], ignore_index=True)


def get_object_version(
    aggr_id: int,
    version: Optional[int] = None,
    model_type: Union[Literal["model", "executor", "dataset_loader"], ModelType] = ModelType.MODEL,
) -> ModelVersionInfo | ExecutorVersionInfo | DatasetLoaderVersionInfo:
    """
    Meta information about the version by the aggr_id and version.

    Parameters
    ----------
    aggr_id: int
        Id of the object.
    version: Optional[int] = None
        Version of the object. Default: None, "latest" version is used.
    model_type: Union[Literal['model', 'executor', 'dataset_loader'] , ModelType]
        Type of the entity. Possible values: ModelType.MODEL | ModelType.EXECUTOR | ModelType.DATASET_LOADER

    Returns
    -------
    ModelVersionInfo|ExecutorVersionInfo|DatasetLoaderVersionInfo
        Instance with meta information.
    """
    model_type = ModelType(model_type)

    op = Operation(schema.Query)

    base_query = _base_get_version(op, aggr_id, version, model_type)

    if model_type == ModelType.MODEL:
        base_query.build_job.status()
        base_query.build_job.build_object_name()
        base_query.available_executor_versions.name()
        base_query.available_executor_versions.version()

    base_query.name()
    base_query.aggr_id()
    base_query.tags()
    base_query.version()
    base_query.experiment.name()
    base_query.experiment.experiment_id()

    model_version = send_graphql_request(op, json_response=False)

    if model_type == ModelType.MODEL:
        return model_version.model_version_from_aggr_id_version

    elif model_type == ModelType.EXECUTOR:
        return model_version.executor_version_from_aggr_id_version

    else:
        return model_version.dataset_loader_version_from_aggr_id_version


def get_object_version_conda_env(
    aggr_id: int,
    version: Optional[int] = None,
    model_type: Union[Literal["model", "executor", "dataset_loader"], ModelType] = ModelType.MODEL,
) -> dict:
    """
    Condas configuration for the object version by the aggr_id and version.

    Parameters
    ----------
    aggr_id: int
        Id of the object.
    version: Optional[int] = None
        Version of the object. Default: None, "latest" version is used.
    model_type: Union[Literal['model', 'executor', 'dataset_loader'] , ModelType]
        Type of the entity. Possible values: ModelType.MODEL | ModelType.EXECUTOR | ModelType.DATASET_LOADER

    Returns
    -------
    Dict
        Dict with conda configuration.
    """
    model_type = ModelType(model_type)

    op = Operation(schema.Query)

    base_query = _base_get_version(op, aggr_id, version, model_type)
    base_query.get_conda_env()
    model_version = send_graphql_request(op, json_response=False)

    if model_type == ModelType.MODEL:
        return model_version.model_version_from_aggr_id_version.get_conda_env

    elif model_type == ModelType.EXECUTOR:
        return model_version.executor_version_from_aggr_id_version.get_conda_env

    else:
        return model_version.dataset_loader_version_from_aggr_id_version.get_conda_env


def get_object_version_requirements(
    aggr_id: int,
    version: int,
    model_type: Union[Literal["model", "executor", "dataset_loader"], ModelType] = ModelType.MODEL,
) -> list:
    """
    Requirements for the object version by the aggr_id and version.

    Parameters
    ----------
    aggr_id: int
        Id of the object.
    version: Optional[int] = None
        Version of the object. Default: None, "latest" version is used.
    model_type: Union[Literal['model', 'executor', 'dataset_loader'] , ModelType]
        Type of the entity. Possible values: ModelType.MODEL | ModelType.EXECUTOR | ModelType.DATASET_LOADER

    Returns
    -------
    List
        List of requirements.
    """
    op = Operation(schema.Query)
    base_query = _base_get_version(op, aggr_id, version, model_type)

    base_query.list_requirements()
    model_version = send_graphql_request(op, json_response=False)

    if model_type == ModelType.MODEL:
        return model_version.model_version_from_aggr_id_version.list_requirements

    elif model_type == ModelType.EXECUTOR:
        return model_version.executor_version_from_aggr_id_version.list_requirements

    else:
        return model_version.dataset_loader_version_from_aggr_id_version.list_requirements


def get_initial_object_version(
    aggr_id: int, model_type: Union[Literal["model", "executor", "dataset_loader"], ModelType] = ModelType.MODEL
) -> ModelVersionInfo | ExecutorVersionInfo | DatasetLoaderVersionInfo:
    """
    Initial object version by the aggr_id.

    Parameters
    ----------
    aggr_id: int
        Id of the object.
    model_type: Union[Literal['model', 'executor', 'dataset_loader'] , ModelType]
        Type of the entity. Possible values: ModelType.MODEL | ModelType.EXECUTOR | ModelType.DATASET_LOADER

    Returns
    -------
    ModelVersionInfo|ExecutorVersionInfo|DatasetLoaderVersionInfo
        Instance with meta information.
    """
    model_type = ModelType(model_type)

    op = Operation(schema.Query)

    if model_type == ModelType.MODEL:
        base_query = op.model_from_id(aggr_id=aggr_id).init_model_version()

    elif model_type == ModelType.EXECUTOR:
        base_query = op.executor_from_id(aggr_id=aggr_id).init_executor_version()

    else:
        base_query = op.dataset_loader_from_id(aggr_id=aggr_id).init_dataset_loader_version()
    version = base_query
    version.name()
    version.version()
    version.tags()
    version.description()

    if model_type == ModelType.MODEL:
        model_version = send_graphql_request(op, json_response=False).model_from_id.init_model_version

    elif model_type == ModelType.EXECUTOR:
        model_version = send_graphql_request(op, json_response=False).executor_from_id.init_executor_version

    else:
        model_version = send_graphql_request(op, json_response=False).dataset_loader_from_id.init_dataset_loader_version

    return model_version


def set_object_tags(
    aggr_id: int,
    key: str,
    values: list[str],
    model_type: Union[Literal["model", "executor", "dataset_loader"], ModelType] = ModelType.MODEL,
) -> Union[ModelInfo, DatasetLoaderInfo, ExecutorInfo]:
    """
    Set object tags.

    Parameters
    ----------
    aggr_id: int
        Id of the object.
    model_type: Union[Literal['model', 'executor', 'dataset_loader'] , ModelType]
        Type of the entity. Possible values: ModelType.MODEL | ModelType.EXECUTOR | ModelType.DATASET_LOADER
    key: str
        Key tag.
    values: list[str]
        Value tags.

    Returns
    -------
    ModelInfo|DatasetLoaderInfo|ExecutorInfo
        Instance with meta information.
    """
    model_type = ModelType(model_type)
    op = Operation(schema.Mutation)
    set_tag = op.set_object_tags(aggr_id=aggr_id, key=key, values=values, model_type=model_type.name).__as__(
        _object_map[model_type]
    )
    _entity(set_tag)
    object_tags = send_graphql_request(op=op, json_response=False).set_object_tags
    return object_tags


def reset_object_tags(
    aggr_id: int,
    key: str,
    values: list[str],
    model_type: Union[Literal["model", "executor", "dataset_loader"], ModelType] = ModelType.MODEL,
    new_key: Optional[str] = None,
) -> Union[ModelInfo, DatasetLoaderInfo, ExecutorInfo]:
    """
    Reset object tags.

    Parameters
    ----------
    aggr_id: int
        Id of the object.
    model_type: Union[Literal['model', 'executor', 'dataset_loader'] , ModelType]
        Type of the entity. Possible values: ModelType.MODEL | ModelType.EXECUTOR | ModelType.DATASET_LOADER
    key: str
        Key tag.
    values: list[str]
        Value tag.
    new_key: Optional[str] = None
        New key of a tag.


    Returns
    -------
    ModelInfo|DatasetLoaderInfo|ExecutorInfo
        Instance with meta information.
    """
    model_type = ModelType(model_type)
    op = Operation(schema.Mutation)
    set_tag = op.reset_object_tags(
        aggr_id=aggr_id, key=key, values=values, new_key=new_key, model_type=model_type.name
    ).__as__(_object_map[model_type])
    _entity(set_tag)
    object_tags = send_graphql_request(op=op, json_response=False).reset_object_tags
    return object_tags


def delete_object_tag(
    aggr_id: int,
    key: str,
    model_type: Union[Literal["model", "executor", "dataset_loader"], ModelType] = ModelType.MODEL,
    value: Optional[str] = None,
) -> Union[ModelInfo, DatasetLoaderInfo, ExecutorInfo]:
    """
    Delete object tag.

    Parameters
    ----------
    aggr_id: int
        Id of the object.
    model_type: Union[Literal['model', 'executor', 'dataset_loader'] , ModelType]
        Type of the entity. Possible values: ModelType.MODEL | ModelType.EXECUTOR | ModelType.DATASET_LOADER
    key: str
        Key tag.
    value: Optional[str]=None
        value tag.
    Returns
    -------
    ModelInfo|DatasetLoaderInfo|ExecutorInfo
        Instance with meta information.
    """
    model_type = ModelType(model_type)
    op = Operation(schema.Mutation)
    delete_tag = op.delete_object_tag(aggr_id=aggr_id, key=key, value=value, model_type=model_type.name).__as__(
        _object_map[model_type]
    )
    _entity(delete_tag)
    object_tag = send_graphql_request(op=op, json_response=False).delete_object_tag
    return object_tag


def set_object_description(
    aggr_id: int,
    description: str,
    model_type: Union[Literal["model", "executor", "dataset_loader"], ModelType] = ModelType.MODEL,
) -> Union[ModelInfo, DatasetLoaderInfo, ExecutorInfo]:
    """
    Set object description.

    Parameters
    ----------
    aggr_id: int
        Id of the object.
    model_type: Union[Literal['model', 'executor', 'dataset_loader'] , ModelType]
        Type of the entity. Possible values: ModelType.MODEL | ModelType.EXECUTOR | ModelType.DATASET_LOADER
    description: str
        Description model.

    Returns
    -------
    ModelInfo|DatasetLoaderInfo|ExecutorInfo
        Instance with meta information.
    """
    model_type = ModelType(model_type)
    op = Operation(schema.Mutation)
    set_description = op.update_object(
        aggr_id=aggr_id,
        update_object_form=UpdateObjectForm(new_description=description),
        model_type=model_type.name,
    ).__as__(_object_map[model_type])
    _entity(set_description)

    update_object = send_graphql_request(op=op, json_response=False).update_object
    return update_object


def set_object_visibility(
    aggr_id: int,
    visibility: Union[Literal["private", "public"], VisibilityOptions],
    model_type: Union[Literal["model", "executor", "dataset_loader"], ModelType] = ModelType.MODEL,
) -> Union[ModelInfo, DatasetLoaderInfo, ExecutorInfo]:
    """
    Set object visibility.

    Parameters
    ----------
    aggr_id: int
        Id of the object.
    model_type: Union[Literal['model', 'executor', 'dataset_loader'] , ModelType]
        Type of the entity. Possible values: ModelType.MODEL | ModelType.EXECUTOR | ModelType.DATASET_LOADER
    visibility: Union[Literal['private', 'public'], VisibilityOptions]
        visibility model.

    Returns
    -------
    ModelInfo|DatasetLoaderInfo|ExecutorInfo
        Instance with meta information.
    """
    model_type = ModelType(model_type)

    op = Operation(schema.Mutation)
    set_visibility = op.update_object(
        aggr_id=aggr_id,
        model_type=model_type.name,
        update_object_form=UpdateObjectForm(new_visibility=VisibilityOptions(visibility).name),
    ).__as__(_object_map[model_type])
    _entity(set_visibility)

    update_object = send_graphql_request(op=op, json_response=False).update_object
    return update_object


def rename_object(
    aggr_id: int,
    new_name: str,
    model_type: Union[Literal["model", "executor", "dataset_loader"], ModelType] = ModelType.MODEL,
) -> Union[ModelInfo, DatasetLoaderInfo, ExecutorInfo]:
    """
    Rename object.

    Parameters
    ----------
    aggr_id: int
        Id of the object.
    model_type: Union[Literal['model', 'executor', 'dataset_loader'] , ModelType]
        Type of the entity. Possible values: ModelType.MODEL | ModelType.EXECUTOR | ModelType.DATASET_LOADER
    new_name: str
        new_name object.

    Returns
    -------
    ModelInfo|DatasetLoaderInfo|ExecutorInfo
        Instance with meta information.
    """
    model_type = ModelType(model_type)

    op = Operation(schema.Mutation)
    set_visibility = op.update_object(
        aggr_id=aggr_id,
        model_type=model_type.name,
        update_object_form=UpdateObjectForm(new_name=new_name),
    ).__as__(_object_map[model_type])
    _entity(set_visibility)

    update_object = send_graphql_request(op=op, json_response=False).update_object
    return update_object


def set_object_version_description(
    aggr_id: int,
    version: int,
    description: str,
    model_type: Union[Literal["model", "executor", "dataset_loader"], ModelType] = ModelType.MODEL,
) -> Union[ModelVersionInfo, ExecutorVersionInfo, DatasetLoaderVersionInfo]:
    """
    Set object version description.

    Parameters
    ----------
    aggr_id: int
        Id of the object.
    version: int
        Version of the object.
    model_type: Union[Literal['model', 'executor', 'dataset_loader'] , ModelType]
        Type of the entity. Possible values: ModelType.MODEL | ModelType.EXECUTOR | ModelType.DATASET_LOADER

    description: str
        Description model version.

    Returns
    -------
    Union[ ModelVersionInfo , ExecutorVersionInfo , DatasetLoaderVersionInfo]:
        Instance with meta information.
    """
    model_type = ModelType(model_type)

    op = Operation(schema.Mutation)
    choice = schema.ObjectIdVersionInput(aggr_id=aggr_id, version=version)
    set_description = op.update_object_version(
        object_version=choice,
        update_object_version_form=UpdateObjectVersionForm(new_description=description),
        model_type=model_type.name,
    ).__as__(_object_version_map[model_type])
    set_description.name()
    set_description.version()
    set_description.description()

    update_object = send_graphql_request(op=op, json_response=False).update_object_version
    return update_object


def set_object_version_visibility(
    aggr_id: int,
    version: int,
    visibility: Union[Literal["private", "public"], VisibilityOptions],
    model_type: Union[Literal["model", "executor", "dataset_loader"], ModelType] = ModelType.MODEL,
) -> Union[ModelVersionInfo, ExecutorVersionInfo, DatasetLoaderVersionInfo]:
    """
    Set object version visibility.

    Parameters
    ----------
    aggr_id: int
        Id of the object.
    version: int
        Version of the object.
    model_type: Union[Literal['model', 'executor', 'dataset_loader'] , ModelType]
        Type of the entity. Possible values: ModelType.MODEL | ModelType.EXECUTOR | ModelType.DATASET_LOADER
    visibility: Union[Literal['private', 'public'], VisibilityOptions]
        visibility model version.

    Returns
    -------
    Union[ ModelVersionInfo , ExecutorVersionInfo , DatasetLoaderVersionInfo]
        Instance with meta information.
    """
    model_type = ModelType(model_type)

    op = Operation(schema.Mutation)
    choice = schema.ObjectIdVersionInput(aggr_id=aggr_id, version=version)
    set_visibility = op.update_object_version(
        object_version=choice,
        update_object_version_form=UpdateObjectVersionForm(new_visibility=VisibilityOptions(visibility).name),
        model_type=model_type.name,
    ).__as__(_object_version_map[model_type])
    set_visibility.name()
    set_visibility.version()
    set_visibility.visibility()

    model = send_graphql_request(op=op, json_response=False).update_object_version
    return model


def set_object_version_tags(
    aggr_id: int,
    version: int,
    key: str,
    values: list[str],
    model_type: Union[Literal["model", "executor", "dataset_loader"], ModelType] = ModelType.MODEL,
) -> Union[ModelVersionInfo, ExecutorVersionInfo, DatasetLoaderVersionInfo]:
    """
    Set object version tags.

    Parameters
    ----------
    aggr_id: int
        Id of the object.
    version: int
        Version of the object.
    model_type: Union[Literal['model', 'executor', 'dataset_loader'] , ModelType]
        Type of the entity. Possible values: ModelType.MODEL | ModelType.EXECUTOR | ModelType.DATASET_LOADER
    key: str
        Key tag.
    values: list[str]
        Value tag.

    Returns
    -------
    Union[ ModelVersionInfo , ExecutorVersionInfo , DatasetLoaderVersionInfo]
        Instance with meta information.
    """
    model_type = ModelType(model_type)

    op = Operation(schema.Mutation)
    choice = schema.ObjectIdVersionInput(aggr_id=aggr_id, version=version)
    set_tag = op.set_object_version_tags(
        object_version=choice, key=key, values=values, model_type=model_type.name
    ).__as__(_object_version_map[model_type])
    set_tag.name()
    set_tag.version()
    set_tag.tags()
    model = send_graphql_request(op=op, json_response=False).set_object_version_tags
    return model


def reset_object_version_tags(
    aggr_id: int,
    version: int,
    key: str,
    values: list[str],
    model_type: Union[Literal["model", "executor", "dataset_loader"], ModelType] = ModelType.MODEL,
    new_key: Optional[str] = None,
) -> Union[ModelVersionInfo, ExecutorVersionInfo, DatasetLoaderVersionInfo]:
    """
    Reset object version tags.

    Parameters
    ----------
    aggr_id: int
        Id of the object.
    version: int
        Version of the object.
    model_type: Union[Literal['model', 'executor', 'dataset_loader'] , ModelType]
        Type of the entity. Possible values: ModelType.MODEL | ModelType.EXECUTOR | ModelType.DATASET_LOADER
    key: str
        Key tag.
    values: list[str]
        Value tag.
    new_key: Optional[str] = None
        New key of a tag.

    Returns
    -------
    Union[ ModelVersionInfo , ExecutorVersionInfo , DatasetLoaderVersionInfo]
        Instance with meta information.
    """
    model_type = ModelType(model_type)

    op = Operation(schema.Mutation)
    choice = schema.ObjectIdVersionInput(aggr_id=aggr_id, version=version)
    set_tag = op.reset_object_version_tags(
        object_version=choice, key=key, values=values, new_key=new_key, model_type=model_type.name
    ).__as__(_object_version_map[model_type])
    set_tag.name()
    set_tag.version()
    set_tag.tags()
    model = send_graphql_request(op=op, json_response=False).reset_object_version_tags
    return model


def delete_object_version_tag(
    aggr_id: int,
    version: int,
    key: str,
    model_type: Union[Literal["model", "executor", "dataset_loader"], ModelType] = ModelType.MODEL,
    value: Optional[str] = None,
) -> Union[ModelVersionInfo, ExecutorVersionInfo, DatasetLoaderVersionInfo]:
    """
    Delete object version tag.

    Parameters
    ----------
    aggr_id: int
        Id of the object.
    version: int
        Version of the object.
    model_type: Union[Literal['model', 'executor', 'dataset_loader'] , ModelType]
        Type of the entity. Possible values: ModelType.MODEL | ModelType.EXECUTOR | ModelType.DATASET_LOADER
    key: str
        Key tag.
    value: Optional[str]=None
        value tag.

    Returns
    -------
    ModelVersion
        Instance with meta information.
    """
    model_type = ModelType(model_type)

    op = Operation(schema.Mutation)
    choice = schema.ObjectIdVersionInput(aggr_id=aggr_id, version=version)
    delete_tag = op.delete_object_version_tag(
        object_version=choice, key=key, value=value, model_type=model_type.name
    ).__as__(_object_version_map[model_type])
    delete_tag.name()
    delete_tag.version()
    delete_tag.tags()
    model = send_graphql_request(op=op, json_response=False).delete_object_version_tag
    return model


def delete_object(
    aggr_id: int, model_type: Union[Literal["model", "executor", "dataset_loader"], ModelType] = ModelType.MODEL
) -> bool:
    """
    Delete object and all of it's versions.

    Parameters
    ----------
    aggr_id: int
        Name of the object to delete.
    model_type: Union[Literal['model', 'executor', 'dataset_loader'] , ModelType]
        Type of the entity. Possible values: ModelType.MODEL | ModelType.EXECUTOR | ModelType.DATASET_LOADER
    Returns
    -------
    bool
        Operation success status.
    """
    model_type = ModelType(model_type)

    op = Operation(schema.Mutation)
    op.delete_objects(aggr_ids=[aggr_id], model_type=model_type.name)
    return send_graphql_request(op)["deleteObjects"]


def delete_object_version(
    aggr_id: int,
    version: int,
    model_type: Union[Literal["model", "executor", "dataset_loader"], ModelType] = ModelType.MODEL,
):
    """
    Delete version of a object.

    Parameters
    ----------
    aggr_id: int
        The name of the object.
    version: int
        The version of the object.
    model_type: Union[Literal['model', 'executor', 'dataset_loader'] , ModelType]
        Type of the entity. Possible values: ModelType.MODEL | ModelType.EXECUTOR | ModelType.DATASET_LOADER
    Returns
    -------
    None
    """
    model_type = ModelType(model_type)

    op = Operation(schema.Mutation)
    op.delete_object_versions(aggr_id=aggr_id, versions=[version], model_type=model_type.name).__as__(
        _object_map[model_type]
    ).name()
    send_graphql_request(op, json_response=False)



from ML_management.sdk import (
    dataset_loader,
    executor,
    experiment,
    job,
    job_code,
    local,
    metric,
    model,
    object,
    parameters,
)

list_model = model.list_model
list_dataset_loader = dataset_loader.list_dataset_loader
list_executor = executor.list_executor
add_ml_job = job.add_ml_job
get_available_resources = job.get_available_resources
list_job_by_name = job.list_job_by_name
job_metric_by_id = job.job_metric_by_id
rename_job = job.rename_job
list_model_version = model.list_model_version
list_dataset_loader_version = dataset_loader.list_dataset_loader_version
list_executor_version = executor.list_executor_version
get_model_version = model.get_model_version
rebuild_model_version_image = model.rebuild_model_version_image
print_model_schema_for_executor = executor.print_model_schema_for_executor
print_dataset_loader_schema = dataset_loader.print_dataset_loader_schema
print_executor_schema = executor.print_executor_schema
print_executor_roles = executor.print_executor_roles
get_logs = job.get_logs
get_build_logs = job.get_build_logs
get_venv_build_logs = job.get_venv_build_logs
serve_model = model.serve_model
stop_model_serving = model.stop_model_serving
check_inference_model_readiness = model.check_inference_model_readiness
get_required_classes_by_executor = executor.get_required_model_classes_by_executor
available_metrics = job.available_metrics
metric_history = job.metric_history
list_params_job = job.list_params_job
delete_model_version = model.delete_model_version
delete_model = model.delete_model
delete_executor_version = executor.delete_executor_version
delete_executor = executor.delete_executor
delete_dataset_loader = dataset_loader.delete_dataset_loader
delete_dataset_loader_version = dataset_loader.delete_dataset_loader_version
cancel_job = job.cancel_job
cancel_build_job_for_model_version = model.cancel_build_job_for_model_version
cancel_build_job_for_executor_version = executor.cancel_build_job_for_executor_version
cancel_venv_build_job_for_model_version = model.cancel_venv_build_job_for_model_version
set_experiment_tags = experiment.set_experiment_tags
reset_experiment_tags = experiment.reset_experiment_tags
set_experiment_description = experiment.set_experiment_description
delete_experiment_tag = experiment.delete_experiment_tag
get_experiment_by_id = experiment.get_experiment_by_id
get_experiment_by_name = experiment.get_experiment_by_name
create_experiment = experiment.create_experiment
ModelMethodParams = parameters.ModelMethodParams
ModelVersionChoice = parameters.ModelVersionChoice
ModelForm = parameters.ModelForm
ModelWithRole = parameters.ModelWithRole
AnyModelForm = parameters.AnyModelForm
DatasetLoaderMethodParams = parameters.DatasetLoaderMethodParams
DatasetLoaderForm = parameters.DatasetLoaderForm
DatasetLoaderWithRole = parameters.DatasetLoaderWithRole
AnyDatasetLoaderForm = parameters.AnyDatasetLoaderForm
ResourcesForm = parameters.ResourcesForm
UploadOneNewModelForm = parameters.UploadOneNewModelForm
UploadOneNewModelWithRole = parameters.UploadOneNewModelWithRole
UploadAnyNewModelsForm = parameters.UploadAnyNewModelsForm
get_dataset_loader_version_requirements = dataset_loader.get_dataset_loader_version_requirements
get_executor_version_requirements = executor.get_executor_version_requirements
get_model_version_requirements = model.get_model_version_requirements
get_dataset_loader_version_conda_env = dataset_loader.get_dataset_loader_version_conda_env
get_executor_version_conda_env = executor.get_executor_version_conda_env
get_executor_version = executor.get_executor_version
get_model_version_conda_env = model.get_model_version_conda_env
get_dataset_loader_version = dataset_loader.get_dataset_loader_version
delete_dataset_loader_version_tag = dataset_loader.delete_dataset_loader_version_tag
delete_executor_version_tag = executor.delete_executor_version_tag
delete_model_version_tag = model.delete_model_version_tag
set_dataset_loader_version_tags = dataset_loader.set_dataset_loader_version_tags
reset_dataset_loader_version_tags = dataset_loader.reset_dataset_loader_version_tags
set_executor_version_tags = executor.set_executor_version_tags
reset_executor_version_tags = executor.reset_executor_version_tags
set_model_version_tags = model.set_model_version_tags
reset_model_version_tags = model.reset_model_version_tags
set_dataset_loader_version_description = dataset_loader.set_dataset_loader_version_description
set_executor_version_description = executor.set_executor_version_description
set_model_version_description = model.set_model_version_description
set_model_description = model.set_model_description
set_executor_description = executor.set_executor_description
set_dataset_loader_description = dataset_loader.set_dataset_loader_description
delete_model_tag = model.delete_model_tag
delete_executor_tag = executor.delete_executor_tag
delete_dataset_loader_tag = dataset_loader.delete_dataset_loader_tag
set_dataset_loader_tags = dataset_loader.set_dataset_loader_tags
reset_dataset_loader_tags = dataset_loader.reset_dataset_loader_tags
set_executor_tags = executor.set_executor_tags
reset_executor_tags = executor.reset_executor_tags
set_model_tags = model.set_model_tags
reset_model_tags = model.reset_model_tags
get_latest_model_version = model.get_latest_model_version
get_latest_executor_version = executor.get_latest_executor_version
get_latest_dataset_loader_version = dataset_loader.get_latest_dataset_loader_version
get_initial_model_version = model.get_initial_model_version
get_initial_executor_version = executor.get_initial_executor_version
get_initial_dataset_loader_version = dataset_loader.get_initial_dataset_loader_version

set_dataset_loader_version_visibility = dataset_loader.set_dataset_loader_version_visibility
set_executor_version_visibility = executor.set_executor_version_visibility
set_model_version_visibility = model.set_model_version_visibility
set_model_visibility = model.set_model_visibility
set_executor_visibility = executor.set_executor_visibility
set_dataset_loader_visibility = dataset_loader.set_dataset_loader_visibility
set_experiment_visibility = experiment.set_experiment_visibility
set_job_visibility = job.set_job_visibility
delete_finished_job = job.delete_finished_job
job_from_id = job.job_from_id
job_resource_usage = job.job_resource_usage

get_required_dataset_loader_classes_by_executor = executor.get_required_dataset_loader_classes_by_executor
get_required_model_classes_by_executor = executor.get_required_model_classes_by_executor

pagination_model = model.pagination_model
pagination_model_version = model.pagination_model_version
pagination_model_version_from_id = model.pagination_model_version_from_id
pagination_executor = executor.pagination_executor
pagination_executor_version = executor.pagination_executor_version
pagination_executor_version_from_id = executor.pagination_executor_version_from_id
pagination_dataset_loader = dataset_loader.pagination_dataset_loader
pagination_dataset_loader_version = dataset_loader.pagination_dataset_loader_version
pagination_dataset_loader_version_from_id = dataset_loader.pagination_dataset_loader_version_from_id

pagination_experiment = experiment.pagination_experiment


list_metric_jobs = metric.list_metric_jobs


add_ml_job_local = local.add_ml_job_local
DebugJobResult = local.DebugJobResult
local_prune = local.local_prune


get_model_from_name = model.get_model_from_name
get_executor_from_name = executor.get_executor_from_name
get_dataset_loader_from_name = dataset_loader.get_dataset_loader_from_name


set_object_tags = object.set_object_tags
reset_object_tags = object.reset_object_tags
delete_object_tag = object.delete_object_tag
set_object_description = object.set_object_description
set_object_visibility = object.set_object_visibility
rename_object = object.rename_object
set_object_version_description = object.set_object_version_description
set_object_version_visibility = object.set_object_version_visibility
set_object_version_tags = object.set_object_version_tags
reset_object_version_tags = object.reset_object_version_tags
delete_object_version_tag = object.delete_object_version_tag
delete_object = object.delete_object
delete_object_version = object.delete_object_version
get_object_version = object.get_object_version
get_object_version_conda_env = object.get_object_version_conda_env
get_object_version_requirements = object.get_object_version_requirements
list_object_version = object.list_object_version
get_initial_object_version = object.get_initial_object_version
get_object = object.get_object


add_custom_code_job = job_code.add_custom_code_job
get_available_images = job_code.get_available_images



"""General Job behavior."""
import json
import os
from typing import Optional

from ML_management import mlmanagement, variables
from ML_management.collectors import COLLECTORS
from ML_management.collectors.collector_pattern_to_methods_map import (
    CollectorMethodName,
)
from ML_management.dataset_loader.dataset_loader_pattern_to_methods_map import (
    DatasetLoaderMethodName,
)
from ML_management.executor.executor_pattern_to_methods_map import ExecutorMethodName
from ML_management.graphql.schema import MLMJob
from ML_management.job.params_prepare import method_params_prepare
from ML_management.job.utils import (
    get_collector_params,
    get_dataset_loader_name_version,
    get_dataset_loaders_params,
    get_model_params,
    get_source_models_aggr_id_version,
    prepare_artifacts,
)
from ML_management.mlmanagement.load_api import load_object
from ML_management.mlmanagement.model_type import ModelType
from ML_management.mlmanagement.upload_model_mode import UploadModelMode
from ML_management.mlmanagement.visibility_options import VisibilityOptions
from ML_management.model.model_type_to_methods_map import ModelMethodName


class Job:
    """Class of job."""

    def __init__(
        self,
        logger,
        preload_models: Optional[dict] = None,
        preload_dataset_loaders: Optional[dict] = None,
        preload_executor=None,
    ):
        self.logger = logger
        self.role_model_map = {}
        self.role_dataset_loader_map = {}
        self.executor = None
        if preload_models:
            self.role_model_map = preload_models

        if preload_dataset_loaders:
            self.role_dataset_loader_map = preload_dataset_loaders
        if preload_executor:
            self.executor = preload_executor

    def unset_dataset(self) -> None:
        """Unset dataset for model and executor instance before logging."""
        self.executor.role_dataset_loader_map = {}
        for role in self.role_model_map:
            self.role_model_map[role].dataset = None

    def set_device(self, device) -> None:
        """Set device for executor instance variable."""
        self.executor.device = device

    def set_model_log_args(self, params):
        """Set additional arguments for log_model() to model instance."""
        new_model_names = {}
        for role_params in params.list_role_model_params:
            # if field==null, sgqlc instantiate empty class without attributes
            if not hasattr(role_params.upload_params, "upload_model_mode"):
                continue
            if role_params.upload_params.new_model_name is None:
                continue
            new_model_names[role_params.role] = role_params.upload_params.new_model_name
        for role, name_version in self.source_models_name_version.items():
            self.role_model_map[role].source_model_aggr_id = name_version["aggr_id"]
            self.role_model_map[role].source_model_version = name_version["version"]
            self.role_model_map[role].source_executor_aggr_id = self.executor_aggr_id
            self.role_model_map[role].source_executor_version = self.executor_version
            self.role_model_map[role].source_executor_role = role
            self.role_model_map[role].model_name = (
                name_version["name"] if role not in new_model_names else new_model_names[role]
            )

    @staticmethod
    def get_init_parameters_from_dict(params):
        """Get __init__ params from list of method_params."""
        for param in params:
            if param["method_name"] == ModelMethodName.init.value:
                return param["method_params"]
        return {}

    def load_models(self) -> None:
        if self.role_model_map:
            return
        for role, model in self.source_models_name_version.items():
            self.logger.info(f"Start to load model aggr_id: {model['aggr_id']} version: {model['version']}")
            self.role_model_map[role] = load_object(
                model["aggr_id"],
                model["version"],
                ModelType.MODEL,
                False,
                f"model-{role}",
                self.get_init_parameters_from_dict(self.model_params[role]["model_methods_params"]),
            ).loaded_object
            self.logger.info(
                f"The model aggr_id: {model['aggr_id']} version: {model['version']} was loaded successfully"
            )

    def load_dataset_loaders(self) -> None:
        if self.role_dataset_loader_map:
            return
        for role, dataset_loader in self.dataset_loader_name_version.items():
            self.logger.info(
                f"Start to load datasetloader aggr_id: {dataset_loader['aggr_id']} version: {dataset_loader['version']}"
            )
            self.role_dataset_loader_map[role] = load_object(
                int(dataset_loader["aggr_id"]),
                int(dataset_loader["version"]),
                ModelType.DATASET_LOADER,
                False,
                kwargs_for_init=self.get_init_parameters_from_dict(
                    self.dataset_loader_params[role]["dataset_loader_methods_params"]
                ),
            ).loaded_object
            self.logger.info(
                f"The dataset loader aggr_id: {dataset_loader['aggr_id']} version: {dataset_loader['version']} "
                f"was loaded successfully"
            )

    def setup_dataset_loader_method_parameters(self) -> None:
        self.executor.dataset_loader_method_parameters_dict = {}
        for role, dl_params in self.dataset_loader_params.items():
            self.executor.dataset_loader_method_parameters_dict[role] = {
                DatasetLoaderMethodName(p["method_name"]): method_params_prepare(
                    self.role_dataset_loader_map[role],
                    p["method_name"],
                    p["method_params"],
                )
                for p in dl_params["dataset_loader_methods_params"]
            }

    def setup_models_parameters(self) -> None:
        self.executor.model_method_parameters_dict = {}
        for role, m_params in self.model_params.items():
            self.executor.model_method_parameters_dict[role] = {
                ModelMethodName(p["method_name"]): method_params_prepare(
                    self.role_model_map[role], p["method_name"], p["method_params"]
                )
                for p in m_params["model_methods_params"]
            }

    def load_entities(self, params) -> None:
        self.executor_aggr_id: int = params.executor_params.executor_version_choice.aggr_id
        self.executor_version: int = params.executor_params.executor_version_choice.version
        self.list_role_model_method_params: list = params.list_role_model_params
        self.list_role_data_params: list = params.list_role_data_params

        # MODELS
        self.model_params = get_model_params(self.list_role_model_method_params)
        self.source_models_name_version = get_source_models_aggr_id_version(self.list_role_model_method_params)
        self.load_models()

        # EXECUTOR
        if not self.executor:
            self.logger.info(
                f"Start to load executor aggr_id: {self.executor_aggr_id} version: {self.executor_version}"
            )
            self.executor = load_object(
                self.executor_aggr_id, self.executor_version, ModelType.EXECUTOR, False
            ).loaded_object
            self.logger.info("The executor was loaded successfully")

        # COLLECTORS
        collectors = {
            data_param_role.role: data_param_role.data_params.collector_name
            for data_param_role in self.list_role_data_params
        }
        self.logger.info("Start to load collectors")
        self.role_collectors_map = {role: COLLECTORS[collector_name]() for role, collector_name in collectors.items()}
        self.logger.info("Collectors was loaded successfully")

        # DATASETLOADERS
        self.dataset_loader_params = get_dataset_loaders_params(self.list_role_data_params)
        self.dataset_loader_name_version = get_dataset_loader_name_version(self.list_role_data_params)
        self.load_dataset_loaders()

    def setup_entities(self, params) -> None:
        if not self.is_distributed:
            self.set_model_log_args(params)

        self.executor.role_model_map = self.role_model_map
        self.executor.role_dataset_loader_map = self.role_dataset_loader_map
        self.set_device("cuda" if params.resources.gpu_number != 0 else "cpu")

        self.setup_dataset_loader_method_parameters()
        self.setup_models_parameters()

    def pull_data(self, params) -> None:
        self.list_role_data_params: list = params.list_role_data_params

        # PULL DATA BY COLLECTORS
        collector_params = get_collector_params(self.list_role_data_params)
        collector_method_name_map = {
            role: collector_param["method_name"] for role, collector_param in collector_params.items()
        }
        collector_method_params_map = {
            role: json.loads(collector_param["method_params"]) for role, collector_param in collector_params.items()
        }
        for role, collector in self.role_collectors_map.items():
            try:
                self.role_dataset_loader_map[role].data_path = getattr(
                    collector, CollectorMethodName(collector_method_name_map[role]).name
                )(**collector_method_params_map[role])
            except Exception as err:
                self.logger.info(f"Error during data collection of {role} collector.")
                raise err

        # LOAD DATA IN DATASETLOADERS
        self.executor.role_dataset_map = {}
        for role, dataset_loader in self.role_dataset_loader_map.items():
            try:
                self.executor.role_dataset_map[role] = dataset_loader.get_dataset(
                    **self.executor.dataset_loader_method_parameters_dict[role][DatasetLoaderMethodName.get_dataset]
                )
            except Exception as err:
                self.logger.info(f"Error during data loading by {role} dataset loader.")
                raise err

    def set_params_for_distributed_tasks(self) -> None:
        self.executor.nodes = os.getenv("SLURM_JOB_NODELIST", "").split(",")
        self.executor.master_addr = os.getenv("MASTER_ADDR")
        self.executor.master_port = int(os.getenv("MASTER_PORT", "0"))
        self.executor.node_rank = int(os.getenv("LOCAL_RANK", "-1"))

    def run_job(self, job: MLMJob):
        """General protocol."""
        experiment_name = job.experiment.name
        self.is_distributed = job.params.is_distributed if job.params else False
        params = job.params
        self.load_entities(params=params)
        self.setup_entities(params=params)
        self.pull_data(params=params)
        if self.is_distributed:
            self.set_params_for_distributed_tasks()

        self.logger.info("Start executor methods for job ...")

        try:
            executor_method_name = params.executor_params.executor_method_params.method_name
            executor_method_params = method_params_prepare(
                self.executor,
                ExecutorMethodName(executor_method_name).name,
                json.loads(params.executor_params.executor_method_params.method_params),
            )
            artifacts = getattr(self.executor, ExecutorMethodName(executor_method_name).name)(**executor_method_params)
        except Exception as err:
            self.logger.error(f"Error during execution of the job with obj_uuid {variables.get_secret_uuid()}.")
            raise err

        if self.is_distributed:
            return
        try:
            self.unset_dataset()
            for role_params in params.list_role_model_params:
                # if field==null, sgqlc instantiate empty class without attributes
                if not hasattr(role_params.upload_params, "upload_model_mode"):
                    continue
                role = role_params.role
                upload_model_mode = role_params.upload_params.upload_model_mode
                prepare_new_model_inference = role_params.upload_params.prepare_new_model_inference
                description = role_params.upload_params.description
                start_build_new_model_image = role_params.upload_params.start_build_new_model_image
                new_model_visibility = role_params.upload_params.new_model_visibility
                registered_model_name = (
                    role_params.upload_params.new_model_name
                    if upload_model_mode == UploadModelMode.new_model
                    else self.source_models_name_version[role]["name"]
                )
                if not isinstance(registered_model_name, str):
                    raise RuntimeError(registered_model_name)
                self.logger.info(f"Start uploading model {registered_model_name} with role {role}")

                if artifacts is not None:
                    prepare_artifacts(artifacts, role, self.role_model_map, registered_model_name)

                model_path = os.path.dirname(self.role_model_map[role].artifacts)
                # child model needs to have same additional local dependencies as a parent
                # in current folder structure we know that each folder beside model folder
                # is folder with additional local dependencies
                additional_local_dependencies = os.listdir(os.path.dirname(model_path))
                additional_local_dependencies = [
                    os.path.join(os.path.dirname(model_path), folder) for folder in additional_local_dependencies
                ]
                additional_local_dependencies.remove(model_path)
                mlmanagement.log_api._log_object_src(
                    artifact_path="",
                    description=description,
                    registered_model_name=registered_model_name,
                    source_model_aggr_id=self.source_models_name_version[role]["aggr_id"],
                    source_model_version=int(self.source_models_name_version[role]["version"]),
                    source_executor_aggr_id=self.executor_aggr_id,
                    source_executor_version=self.executor_version,
                    source_executor_role=role,
                    upload_model_mode=upload_model_mode,
                    model_path=model_path,
                    create_venv_pack=prepare_new_model_inference,
                    additional_local_packages=additional_local_dependencies,
                    experiment_name=experiment_name,
                    start_build=start_build_new_model_image,
                    visibility=VisibilityOptions[new_model_visibility],
                )
                self.logger.info(f"Model: {registered_model_name}, role: {role} logged successfully.")
        except Exception as err:
            self.logger.error(f"Error during execution of the job with obj_uuid {variables.get_secret_uuid()}.")
            raise err



import warnings

warnings.warn(
    "SkipJsonSchema will be moved to ML_management.jsonschema_inference in future versions.", DeprecationWarning
)


class SkipJsonSchema:
    """
    This class has been moved to ML_management.jsonschema_inference.

        .. deprecated::
    """

    def __class_getitem__(cls, _type):
        return cls()



"""Module with pydantics model representing parameters of sdk functions."""
import json
from typing import List, Literal, Optional, Union

from pydantic import BaseModel, Field, PositiveInt

from ML_management.dataset_loader.dataset_loader_pattern_to_methods_map import DatasetLoaderMethodName
from ML_management.executor import BaseExecutor
from ML_management.mlmanagement.upload_model_mode import UploadModelMode
from ML_management.mlmanagement.visibility_options import VisibilityOptions
from ML_management.model.model_type_to_methods_map import ModelMethodName


class ModelMethodParams(BaseModel):
    """Class for parameters for one model method.

    Attributes
    ----------
    method: ModelMethodName
        method of model.
    params: dict
        parameters for method.
        Example::

            {key1: value1, key2: value2}

    """

    method: ModelMethodName
    params: dict

    def serialize(self) -> dict:
        return {self.method.value: self.params}

    def serialize_gql(self) -> dict:
        return {"methodName": self.method.value, "methodParams": json.dumps(self.params)}


class DatasetLoaderMethodParams(BaseModel):
    """Class for parameters for one dataset loader method.

    Attributes
    ----------
    method: DatasetLoaderMethodName
        method of dataset loader.
    params: dict
        parameters for method.
        Example::

            {key1: value1, key2: value2}

    """

    method: DatasetLoaderMethodName = DatasetLoaderMethodName.get_dataset
    params: dict

    def serialize(self) -> dict:
        return {self.method.value: self.params}

    def serialize_gql(self) -> dict:
        return {"methodName": self.method.value, "methodParams": json.dumps(self.params)}


class ModelVersionChoice(BaseModel):
    """Class for model version choice in add_ml_job.

    Attributes
    ----------
    aggr_id: int
        ID of the model to interact with.
    version: Optional[int] = None
        Version of the model to interact with.
        Default: None, "latest" version is used.
    """

    aggr_id: int
    version: Optional[PositiveInt] = None

    def serialize(self) -> dict:
        return {
            "aggr_id": self.aggr_id,
            "version": self.version,
        }

    def serialize_gql(self):
        return {"version": self.version, "aggrId": self.aggr_id, "name": "null"}


class ModelForm(BaseModel):
    """Class for single model parameters for add_ml_job function.

    Attributes
    ----------
    model_version_choice: ModelVersionChoice
        parameters that determine which model to use.
    params: List[ModelMethodParams] = []
        List of ModelMethodParams with parameters of model methods.
        Example::

            [
                ModelMethodParams(method=ModelMethodName.evaluate_function, params={key1: value1, key2: value2}),
                ModelMethodParams(method=ModelMethodName.finetune_function, params={key1: value1, key2: value2})
            ]

        Default: [].
    """

    model_version_choice: ModelVersionChoice
    params: List[ModelMethodParams] = []

    def serialize(self) -> List[dict]:
        return [
            {
                "role": BaseExecutor.DEFAULT_ROLE,
                "model": self.model_version_choice.serialize(),
                "params": [method_params.serialize() for method_params in self.params],
            }
        ]

    def serialize_gql(self):
        return [
            {
                "modelParams": {
                    "listModelMethodParams": [method_params.serialize_gql() for method_params in self.params],
                    "modelVersionChoice": self.model_version_choice.serialize_gql(),
                },
                "role": BaseExecutor.DEFAULT_ROLE,
            }
        ]


class ModelWithRole(ModelForm):
    """Class for one role model with parameters for add_ml_job function.

    Same as ModelForm except parameter role: str is required.

    Attributes
    ----------
    model_version_choice: ModelVersionChoice
        parameters that determine which model to use.
    params: List[ModelMethodParams] = []
        List of ModelMethodParams with parameters of model methods.
        Example::

            [
                ModelMethodParams(method=ModelMethodName.evaluate_function, params={key1: value1, key2: value2}),
                ModelMethodParams(method=ModelMethodName.finetune_function, params={key1: value1, key2: value2})
            ]

        Default: [].
    new_model_name: Optional[str] = None
        Name of the model to save in case new model is to be created as a result of job execution.
        (regulated by executor.upload_model_mode).
        Default: None.
    new_model_description: Optional[str] = None
        Description of the new model name.
        Default: None.
    prepare_new_model_inference: bool = False
        Start preparing the environment for the inference of the new model.
        Default: False.
    role: str
        Role of the model.
    """

    role: str

    def serialize(self) -> List[dict]:
        result = super().serialize()
        result[0]["role"] = self.role
        return result

    def serialize_gql(self) -> List[dict]:
        result = super().serialize_gql()
        result[0]["role"] = self.role
        return result


class AnyModelForm(BaseModel):
    """Class for definition of arbitrary number of models used in job with their roles.

    Attributes
    ----------
    models: List[ModelWithRole]

        Example (Both instances of ModelWithRole and usual list of dicts may be used)::

            [
                ModelWithRole(
                    role="role_one",
                    model_version_choice=ModelVersionChoice(
                        name="model_1_name",
                        version=1,  # Optional. Default: latest version
                    )
                    params=[
                        ModelMethodParams(
                            method=ModelMethodName.evaluate_function,
                            params={key1: value1, key2: value2}
                        ),
                        ModelMethodParams(
                            method=ModelMethodName.finetune_function,
                            params={key1: value1, key2: value2}
                        )
                    ]
                    new_model_name="some_name"  # Optional. Default: None
                    new_model_description="some_description" # Optional. Default: None
                ),
                ModelWithRole(
                    role="role_two",
                    model_version_choice=ModelVersionChoice(name="model_2_name"),
                    params=[
                        ModelMethodParams(
                            method=ModelMethodName.evaluate_function,
                            params={key1: value1, key2: value2}
                        ),
                        ModelMethodParams(
                            method=ModelMethodName.finetune_function,
                            params={key1: value1, key2: value2}
                        )
                    ]
                )
            ]

    """

    models: List[ModelWithRole]

    def serialize(self) -> List[dict]:
        if len(self.models) == 0:
            return []

        result = self.models[0].serialize()
        for model in self.models[1:]:
            result.extend(model.serialize())
        return result

    def serialize_gql(self) -> List[dict]:
        if len(self.models) == 0:
            return []

        result = self.models[0].serialize_gql()
        for model in self.models[1:]:
            result.extend(model.serialize_gql())
        return result


class DatasetLoaderForm(BaseModel):
    """Class for single dataset loader parameters for add_ml_job function.

    Attributes
    ----------
    aggr_id: int
        ID of the DatasetLoader that the model will use.
    version: Optional[int] = None
        Version of the DatasetLoader that the model will interact with.
        Default: None, "latest" version is used.
    params: List[DatasetLoaderMethodParams] = []
        List of DatasetLoaderMethodParams with parameters of model methods.
        Example::

            [
                DatasetLoaderMethodParams(
                    # by default params are set for DatasetLoaderMethodName.get_dataset method
                    params={key1: value1, key2: value2}
                ),
                DatasetLoaderMethodParams(
                    method=DatasetLoaderMethodName.<some_method>,
                    params={key1: value1, key2: value2}
                )
            ]

        Default: [].
    collector_name: {"s3", }
        Name of the collector to interact with. Default: "s3"
    collector_params: dict
        Dictionary of collector parameters.
        Example::

            {"bucket": "mnist"}

    """

    aggr_id: int
    params: List[DatasetLoaderMethodParams] = []
    collector_params: dict
    version: Optional[PositiveInt] = None
    collector_name: str = "s3"

    def serialize(self) -> List[dict]:
        return [
            {
                "role": BaseExecutor.DEFAULT_ROLE,
                "data_params": {
                    "collector_name": self.collector_name,
                    "collector_params": self.collector_params,
                    "dataset_loader_aggr_id": self.aggr_id,
                    "dataset_loader_version": self.version,
                    "dataset_loader_params": [method_params.serialize() for method_params in self.params],
                },
            }
        ]

    def serialize_gql(self) -> List[dict]:
        return [
            {
                "dataParams": {
                    "collectorMethodParams": {
                        "methodName": "collector_method",
                        "methodParams": json.dumps(self.collector_params),
                    },
                    "collectorName": self.collector_name,
                    "datasetLoaderVersionChoice": {"aggrId": self.aggr_id, "version": self.version},
                    "listDatasetLoaderMethodParams": [method_params.serialize_gql() for method_params in self.params],
                },
                "role": BaseExecutor.DEFAULT_ROLE,
            }
        ]


class DatasetLoaderWithRole(DatasetLoaderForm):
    """Class for one role dataset loader with parameters for add_ml_job function.

    Same as DatasetLoaderForm except parameter role: str is required.

    Attributes
    ----------
    name: str
        Name of the DatasetLoader that the model will use.
    version: Optional[int] = None
        Version of the DatasetLoader that the model will interact with.
        Default: None, "latest" version is used.
    params: List[DatasetLoaderMethodParams] = []
        List of DatasetLoaderMethodParams with parameters of model methods.
        Example::

            [
                DatasetLoaderMethodParams(
                    # by default params are set for DatasetLoaderMethodName.get_dataset method
                    params={key1: value1, key2: value2}
                ),
                DatasetLoaderMethodParams(
                    method=DatasetLoaderMethodName.<some_method>,
                    params={key1: value1, key2: value2}
                )
            ]

        Default: [].
    collector_name: {"s3", }
        Name of the collector to interact with.
        Default: "s3"
    collector_params: dict
        Dictionary of collector parameters.
        Example::

            {"bucket": "mnist"}

    role: str
        Role of the DatasetLoader.
    """

    role: str

    def serialize(self) -> List[dict]:
        result = super().serialize()
        result[0]["role"] = self.role
        return result

    def serialize_gql(self) -> List[dict]:
        result = super().serialize_gql()
        result[0]["role"] = self.role
        return result


class AnyDatasetLoaderForm(BaseModel):
    """Class for definition of arbitrary number of dataset_loaders used in job with their roles.

    Attributes
    ----------
    datasetloaders: List[DatasetLoaderWithRole]

        Example (Both instances of DatasetLoaderWithRole and usual list of dicts may be used)::

            [
                DatasetLoaderWithRole(
                    role="data1",
                    name="multiple_two_datasets",
                    version=1,  # Optional. Default: latest version
                    params=[
                        DatasetLoaderMethodParams(
                            # by default params are set for DatasetLoaderMethodName.get_dataset method
                            params={key1: value1, key2: value2}
                        ),
                        DatasetLoaderMethodParams(
                            method=DatasetLoaderMethodName.get_dataset,  # or other method
                            params={key1: value1, key2: value2}
                        )
                    ],
                    collector_name="s3",
                    collector_params={"bucket": "mnist"}
                ),
                DatasetLoaderWithRole(
                    role="data2",
                    name="multiple_two_datasets",
                    collector_name="s3",
                    collector_params={"bucket": "mnist"}
                )
            ]

    """

    dataset_loaders: List[DatasetLoaderWithRole]

    def serialize(self) -> List[dict]:
        if len(self.dataset_loaders) == 0:
            return []

        result = self.dataset_loaders[0].serialize()
        for model in self.dataset_loaders[1:]:
            result.extend(model.serialize())
        return result

    def serialize_gql(self) -> List[dict]:
        if len(self.dataset_loaders) == 0:
            return []

        result = self.dataset_loaders[0].serialize_gql()
        for model in self.dataset_loaders[1:]:
            result.extend(model.serialize_gql())
        return result


class UploadOneNewModelForm(BaseModel):
    """Class for definition of parameters of one new model to be uploaded.

    Attributes
    ----------
    upload_model_mode: UploadModelMode
        Option how to upload a new model.
    new_model_name: Optional[str] = None
        Name of the model to save in case new model is to be created as a result of job execution.
        (regulated by upload_model_mode).
        Default: None.
    new_model_description: Optional[str] = None
        Description of the new model name.
        Default: None.
    prepare_new_model_inference: bool = False
        Start preparing the environment for the inference of the new model.
        Default: False.
    start_build_new_model_image: bool = True
        Start build image for the new model.
        Default: True.
    new_model_visibility: Union[Literal[VisibilityOptions.PRIVATE, VisibilityOptions.PUBLIC], VisibilityOptions]
        Visibility of the new model to other users. Default: PRIVATE.
    """

    upload_model_mode: UploadModelMode
    new_model_name: Optional[str] = None
    new_model_description: Optional[str] = None
    prepare_new_model_inference: bool = False
    start_build_new_model_image: bool = True
    new_model_visibility: Union[
        Literal[VisibilityOptions.PRIVATE, VisibilityOptions.PUBLIC], VisibilityOptions
    ] = VisibilityOptions.PRIVATE

    def serialize(self) -> dict:
        return {
            BaseExecutor.DEFAULT_ROLE: {
                "upload_model_mode": self.upload_model_mode.name,
                "new_model_name": self.new_model_name,
                "prepare_new_model_inference": self.prepare_new_model_inference,
                "start_build_new_model_image": self.start_build_new_model_image,
                "new_model_visibility": VisibilityOptions(self.new_model_visibility).name,
                "description": self.new_model_description,
            }
        }

    def serialize_gql(self) -> dict:
        return {
            BaseExecutor.DEFAULT_ROLE: {
                "uploadModelMode": self.upload_model_mode.name,
                "newModelName": self.new_model_name,
                "prepareNewModelInference": self.prepare_new_model_inference,
                "startBuildNewModelImage": self.start_build_new_model_image,
                "newModelVisibility": VisibilityOptions(self.new_model_visibility).name,
                "description": self.new_model_description,
            }
        }


class UploadOneNewModelWithRole(UploadOneNewModelForm):
    """Class for definition of parameters of one new model to be uploaded.

    Same as UploadOneNewModelForm except parameter role: str is required.

    Attributes
    ----------
    upload_model_mode: UploadModelMode
        Option how to upload a new model.
    new_model_name: Optional[str] = None
        Name of the model to save in case new model is to be created as a result of job execution.
        (regulated by upload_model_mode).
        Default: None.
    new_model_description: Optional[str] = None
        Description of the new model name.
        Default: None.
    prepare_new_model_inference: bool = False
        Start preparing the environment for the inference of the new model.
        Default: False.
    start_build_new_model_image: bool = True
        Start build image for the new model.
        Default: True.
    new_model_visibility: Union[Literal[VisibilityOptions.PRIVATE, VisibilityOptions.PUBLIC], VisibilityOptions]
        Visibility of the new model to other users. Default: PRIVATE.
    """

    role: str

    def serialize(self) -> dict:
        result = super().serialize()
        result = {self.role: list(result.values())[0]}
        return result

    def serialize_gql(self) -> dict:
        result = super().serialize_gql()
        result = {self.role: list(result.values())[0]}
        return result


class UploadAnyNewModelsForm(BaseModel):
    """Class for definition of parameters new models to be uploaded."""

    upload_models_params: List[UploadOneNewModelWithRole]

    def serialize(self) -> dict:
        if len(self.upload_models_params) == 0:
            return {}

        result = self.upload_models_params[0].serialize()
        for params in self.upload_models_params[1:]:
            result.update(params.serialize())
        if len(self.upload_models_params) != len(result):
            raise ValueError("All roles must be different, you can not pass same role twice.")
        return result

    def serialize_gql(self) -> dict:
        if len(self.upload_models_params) == 0:
            return {}

        result = self.upload_models_params[0].serialize_gql()
        for params in self.upload_models_params[1:]:
            result.update(params.serialize_gql())
        if len(self.upload_models_params) != len(result):
            raise ValueError("All roles must be different, you can not pass same role twice.")
        return result


class ResourcesForm(BaseModel):
    """Class for definition of resources required for job execution.

    The computing cluster distributes all resources equally among the allocated nodes.
    If such an equal distribution is not possible, the job will be rejected.

    Attributes
    ----------
    cpus: PositiveInt
        Number of cpus to use in job. Default: 1.
    memory_per_node: PositiveInt
        Number of memory in GB that is required for job process on one node.
        If job has started successfully this amount of memory is guaranteed for you,
        but exceeding the amount of this memory can lead to OOM-kill result.
        IMPORTANT: some job execution backends may spend this memory for job image unpacking,
        it is always true in case of distributed jobs, because they are supported only by one execution backend.
        Default: 8.
    gpu_number: int
        Number of gpu to be used in job. Default: 0.
    gpu_type: Optional[str]
        Type of gpu to be used in job. If you require specific gpu type, i.e. A100,
        you should specify this parameter. Default: None.
    """

    cpus: PositiveInt = 1
    memory_per_node: PositiveInt = 8
    gpu_number: int = Field(0, ge=0)
    gpu_type: Optional[str] = None



import fcntl
import os
from dataclasses import dataclass
from pathlib import Path


def get_upload_paths(local_path: str):
    """Return all file paths in folder."""
    if os.path.isfile(local_path):
        file_size = os.stat(local_path).st_size  # size of file, in bytes
        return [
            StorageFilePath(storage_path=os.path.basename(local_path), local_path=local_path, size=file_size)
        ], file_size

    local_files = [str(path) for path in Path(local_path).rglob("*") if path.is_file()]

    upload_paths = []
    size = 0
    for local_file_path in local_files:
        storage_file_path = os.path.relpath(local_file_path, local_path)
        file_size = os.stat(local_file_path).st_size  # size of file, in bytes
        upload_paths.append(StorageFilePath(storage_path=storage_file_path, local_path=local_file_path, size=file_size))
        size += file_size

    return upload_paths, size


def get_upload_size(local_path: str):
    """Return size of folder or file."""
    return sum(f.stat().st_size for f in Path(local_path).glob("**/*") if f.is_file())


def get_bucket_size(bucket: str, remote_paths: list, paginator):
    """Return size of bucket."""
    total_bucket_size = 0
    for remote_path in remote_paths:
        page_iterator = paginator.paginate(Bucket=bucket, Prefix=remote_path)
        for page in page_iterator:
            for obj in page.get("Contents", []):
                total_bucket_size += obj["Size"]
    return total_bucket_size


def get_bucket_info(bucket: str, remote_paths: list, paginator):
    """Return size of bucket."""
    total_bucket_size = 0
    bucket_info = {}
    for remote_path in remote_paths:
        page_iterator = paginator.paginate(Bucket=bucket, Prefix=remote_path)
        for page in page_iterator:
            for obj in page.get("Contents", []):
                bucket_info[obj["Key"]] = {"LastModified": obj["LastModified"], "Size": obj["Size"]}
                total_bucket_size += obj["Size"]
    return total_bucket_size, bucket_info


@dataclass
class StorageFilePath:
    """Define paths for file in S3 Storage."""

    local_path: str
    storage_path: str
    size: int

    def __post_init__(self):
        """Check the types of variables."""
        assert isinstance(self.local_path, str)
        assert isinstance(self.storage_path, str)
        assert isinstance(self.size, int)


class FileLock:
    """Lock the synchronization by '.<bucket_name>' file as mutex."""

    def __init__(self, path: str, bucket: str) -> None:
        lock_directory = os.path.join(path, ".lock_files")
        self.lock_filename = os.path.join(lock_directory, "." + bucket)
        os.makedirs(lock_directory, exist_ok=True)

    def __enter__(self):
        self.fd = os.open(self.lock_filename, os.O_RDWR | os.O_CREAT)
        fcntl.lockf(self.fd, fcntl.LOCK_EX)

    def __exit__(self, _type, value, tb):
        fcntl.lockf(self.fd, fcntl.LOCK_UN)
        os.close(self.fd)



"""Enum of existing model types."""
import warnings
from enum import Enum

warnings.warn("ModelType will be moved to ML_management.types.model_type in future versions.", DeprecationWarning)


class ModelType(str, Enum):
    """Model type class."""

    MODEL = "model"
    EXECUTOR = "executor"
    DATASET_LOADER = "dataset_loader"



from collections import defaultdict

from ML_management.singleton_pattern import Singleton


class MetricAutostepper(metaclass=Singleton):
    def __init__(self):
        self._data = defaultdict(int)

    def get_next_step(self, key):
        step = self._data[key]
        self._data[key] += 1
        return step

    def clear(self):
        self._data = defaultdict(int)



"""SDK for client library."""
import json
from typing import Dict, List

import pandas as pd


def _to_datetime(df: pd.DataFrame, column_names: List[str]) -> pd.DataFrame:
    """
    Convert df's columns to datetime.

    Parameters
    ----------
    df: pd.DataFrame
        pd.DataFrame in which the columns will be converted.
    column_names: List[str]
        Column names to be converted.

    Returns
    -------
    pd.DataFrame
        Pandas dataframe with converted columns.
    """
    for column_name in column_names:
        df[column_name] = pd.to_datetime(df[column_name], unit="s")

    return df


def _print_params_by_schema(json_schema: Dict, schema_type: str) -> None:
    """Print entity JSON Schema with required params."""
    properties_and_required_dict = {key: json_schema[key] for key in ("properties", "required") if key in json_schema}

    json_formatted_str = json.dumps(properties_and_required_dict, indent=2)

    print(f"{schema_type} json-schema:")

    print(json_formatted_str)


def _entity(base):
    base.name()
    base.aggr_id()
    base.description()
    base.creation_timestamp()
    base.last_updated_timestamp()
    base.tags()
    base.visibility()



"""Common variables."""
import os
import posixpath
from typing import Optional, Tuple

server_url = None
s3_url = None
s3_username = "PLACEHOLDER"
s3_password = "PLACEHOLDER"
mlm_login = None
mlm_password = None

secret_uuid = None
active_job = False

CONFIG_KEY_ARTIFACTS = "artifacts"
DEFAULT_EXPERIMENT = "Default"
MLCONFIG = "MLConfig.yaml"
MLMODEL_FILE_NAME = "MLmodel"
DATA = "entity_code_data"
LEGACY_DATA = "data"
FLAVOR_NAME = "python_function"
CONDA_SIZE_LIMIT = 50000
INFERENCE_CONFIG_LIMIT = 50000

METRIC_ACCUMULATION_DURATION = None
TIMEOUT_LOG_METRIC_BATCH = None
BUTCH_POLLING_FREQUENCY = None

EXPERIMENT_NAME_FOR_EXECUTOR = "executors"
EXPERIMENT_NAME_FOR_DATASET_LOADER = "dataset_loaders"
FILENAME_FOR_INFERENCE_CONFIG = "predict_config.json"
SERVER_META = posixpath.join("bff", "meta.json")

LOCAL_REGISTRY_PATH = os.path.abspath(
    os.path.expanduser(os.environ.get("LOCAL_REGISTRY_PATH", "~/.mlm_registry_cache"))
)
CACHED_LIST_FILENAME = "list.json"
JOB_ARTIFACT_DIRNAME = "job_artifacts"
NO_CACHE = bool(os.environ.get("NO_CACHE", False))

DEBUG = False
DEBUG_REGISTRY_PATH = os.path.abspath(
    os.path.expanduser(os.environ.get("DEBUG_REGISTRY_PATH", "~/.mlm_debug_registry"))
)


def get_secret_uuid() -> str:
    return secret_uuid if secret_uuid else os.environ.get("secret_uuid", None)


def get_metric_accumulation_duration() -> float:
    return (
        METRIC_ACCUMULATION_DURATION
        if METRIC_ACCUMULATION_DURATION
        else float(os.environ.get("METRIC_ACCUMULATION_DURATION", 1))
    )


def get_timeout_log_metric_batch() -> float:
    return (
        TIMEOUT_LOG_METRIC_BATCH if TIMEOUT_LOG_METRIC_BATCH else float(os.environ.get("TIMEOUT_LOG_METRIC_BATCH", 20))
    )


def get_butch_polling_frequency() -> float:
    return BUTCH_POLLING_FREQUENCY if BUTCH_POLLING_FREQUENCY else float(os.environ.get("BUTCH_POLLING_FREQUENCY", 0.3))


def get_log_service_url(function_name: str) -> str:
    """Get server '/log-object' endpoint URL for log_model, log_artifact, download_artifacts functions."""
    log_object_url = os.environ.get("log_object_url")
    base_url = log_object_url if log_object_url is not None else get_server_url()
    return posixpath.join(base_url, "log-object", function_name.replace("_", "-"))


def get_server_url() -> str:
    """
    Get server URL.

    If you set the URL using 'mlmanagement.set_server_url' function,
    it takes precedence over the URL from the environment variable 'server_url'
    """
    return os.environ.get("server_url", "https://local.tai-dev.intra.ispras.ru") if not server_url else server_url


def get_s3_gateway_url() -> str:
    """
    Get s3 URL.

    If you set the URL using 'mlmanagement.set_s3_url' function,
    it takes precedence over the URL from the environment variable 'S3_URL'
    """
    return os.environ.get("S3_URL", get_server_url()) if not s3_url else s3_url


def get_s3_credentials() -> Tuple[str, str]:
    """Get s3 credentials."""
    return s3_username, s3_password


def get_mlm_credentials() -> Tuple[Optional[str], Optional[str]]:
    """
    Get mlm credentials.

    If you set the URL using 'mlmanagement.set_mlm_credentials' function,
    it takes precedence over the URL from the environment variables 'MLM_LOGIN' and 'MLM_PASSWORD'.
    Environment variables 'login' and 'password' have last priority.
    """
    login = (
        (os.getenv("login") if not os.getenv("MLM_LOGIN") else os.getenv("MLM_LOGIN")) if not mlm_login else mlm_login
    )
    password = (
        (os.getenv("password") if not os.getenv("MLM_PASSWORD") else os.getenv("MLM_PASSWORD"))
        if not mlm_password
        else mlm_password
    )

    return login, password


def get_server_websocket_url() -> str:
    """Get the current websocket server URL."""
    url = get_server_url()
    splitted_url = url.split("/")
    splitted_url[0] = "wss:" if "https" in url else "ws:"
    return "/".join(splitted_url)



from .patterns.evaluatable_model import EvaluatableModel
from .patterns.gradient_model import GradientModel
from .patterns.model_pattern import Model
from .patterns.model_with_losses import ModelWithLosses
from .patterns.model_with_metrics import ModelWithMetrics
from .patterns.preprocessor import Preprocessor
from .patterns.retrainable_model import RetrainableModel
from .patterns.target_layer import TargetLayer
from .patterns.torch_model import TorchModel
from .patterns.trainable_model import TrainableModel
from .patterns.transformer import Transformer

EvaluatableModel = EvaluatableModel
GradientModel = GradientModel
Model = Model
ModelWithLosses = ModelWithLosses
ModelWithMetrics = ModelWithMetrics
Preprocessor = Preprocessor
RetrainableModel = RetrainableModel
TargetLayer = TargetLayer
TorchModel = TorchModel
TrainableModel = TrainableModel
Transformer = Transformer



"""General testing of the sdk."""
import os
import unittest
from typing import Union

import pandas

import ML_management.sdk.dataset_loader
import ML_management.sdk.executor
import ML_management.sdk.job
import ML_management.sdk.model
from ML_management.graphql.schema import DatasetLoaderInfo, ExecutorInfo, MLMJob, ModelInfo, ModelVersionInfo
from ML_management.mlmanagement import set_server_url
from ML_management.sdk.parameters import (
    AnyModelForm,
    DatasetLoaderForm,
    DatasetLoaderMethodParams,
    ModelVersionChoice,
    ModelWithRole,
)

os.environ["kc_access"] = "kc_access"
os.environ["kc_state"] = "kc_state"

NAME = "name"
AGGR_ID = 1
VERSION = 1
PARAMS = {}
ROLE = "single"


class TestSDK(unittest.TestCase):
    """Tests that sdk is supported by current server introspection."""

    def setUp(self):
        """Set mock_server url."""
        set_server_url("http://localhost:4000/")

    def test_list_model(self):
        """Test sdk.list_model method returns correct answer."""
        list_model = ML_management.sdk.model.list_model()
        # mock server generates non-empty result by introspection.
        self.assertGreater(len(list_model), 0)
        # sdk returns an object of the correct type.
        self.assertIsInstance(list_model, pandas.core.frame.DataFrame)

    def test_list_dataset_loader(self):
        """Test sdk.list_dataset_loader method returns correct answer."""
        list_dataset_loader = ML_management.sdk.dataset_loader.list_dataset_loader()
        # mock server generates non-empty result by introspection.
        self.assertGreater(len(list_dataset_loader), 0)
        # sdk returns an object of the correct type.
        self.assertIsInstance(list_dataset_loader, pandas.core.frame.DataFrame)

    def test_list_executor(self):
        """Test sdk.list_executor method returns correct answer."""
        list_executor = ML_management.sdk.executor.list_executor()
        # mock server generates non-empty result by introspection.
        self.assertGreater(len(list_executor), 0)
        # sdk returns an object of the correct type.
        self.assertIsInstance(list_executor, pandas.core.frame.DataFrame)

    def test_add_ml_job(self):
        """Test sdk.add_ml_job method returns correct answer."""
        self.assertIsInstance(
            ML_management.sdk.job.add_ml_job(
                job_executor_aggr_id=AGGR_ID,
                executor_params=PARAMS,
                models_pattern=AnyModelForm(
                    models=[
                        ModelWithRole(
                            role=ROLE, model_version_choice=ModelVersionChoice(aggr_id=AGGR_ID, version=VERSION)
                        )
                    ]
                ),
                data_pattern=DatasetLoaderForm(
                    aggr_id=AGGR_ID,
                    version=VERSION,
                    params=[DatasetLoaderMethodParams(params=PARAMS)],
                    collector_params=PARAMS,
                ),
                job_executor_version=VERSION,
            ),
            MLMJob,
        )

    def test_list_job_by_name(self):
        """Test sdk.job.list_job_by_name method returns correct answer."""
        jobs = ML_management.sdk.job.list_job_by_name("job_name")
        self.assertIsInstance(jobs, list)

    def test_job_metric_by_name(self):
        """Test sdk.job_metric_by_name method returns correct answer."""
        metric = ML_management.sdk.job.job_metric_by_id(job_id=VERSION)
        # mock server generates non-empty result by introspection.
        self.assertEqual(len(metric), 1)
        # sdk returns an object of the correct type.
        self.assertIsInstance(ML_management.sdk.job.job_metric_by_id(job_id=VERSION), pandas.core.frame.DataFrame)

    def test_list_model_version(self):
        """Test sdk.list_model_version method returns correct answer."""
        list_model_version = ML_management.sdk.model.list_model_version(name=NAME)
        # mock server generates non-empty result by introspection.
        self.assertGreater(len(list_model_version), 0)
        # sdk returns an object of the correct type.
        self.assertIsInstance(list_model_version, pandas.core.frame.DataFrame)

    def test_list_dataset_loader_version(self):
        """Test sdk.list_dataset_loader_version method returns correct answer."""
        list_dataset_loader_version = ML_management.sdk.dataset_loader.list_dataset_loader_version(name=NAME)
        # mock server generates non-empty result by introspection.
        self.assertGreater(len(list_dataset_loader_version), 0)
        # sdk returns an object of the correct type.
        self.assertIsInstance(list_dataset_loader_version, pandas.core.frame.DataFrame)

    def test_list_executor_version(self):
        """Test sdk.list_executor_version method returns correct answer."""
        list_executor_version = ML_management.sdk.executor.list_executor_version(name=NAME)
        # mock server generates non-empty result by introspection.
        self.assertGreater(len(list_executor_version), 0)
        # sdk returns an object of the correct type.
        self.assertIsInstance(list_executor_version, pandas.core.frame.DataFrame)

    def test_model_version_metainfo(self):
        """Test sdk.model_version_metainfo method returns correct answer."""
        self.assertIsInstance(ML_management.sdk.model.get_model_version(name=NAME, version=VERSION), ModelVersionInfo)

    def test_rebuild_model_version_image(self):
        """Test sdk.rebuild_model_version_image method returns correct answer."""
        self.assertIsInstance(
            ML_management.sdk.model.rebuild_model_version_image(aggr_id=AGGR_ID, version=VERSION), bool
        )

    def test_available_metrics(self):
        """Test sdk.available_metrics method returns correct answer."""
        self.assertIsInstance(ML_management.sdk.job.available_metrics(VERSION), list)
        self.assertIsInstance(ML_management.sdk.job.available_metrics(VERSION)[0], str)

    def test_get_required_classes_by_executor(self):
        """Test sdk.get_required_classes_by_executor method returns correct answer."""
        required_classes = ML_management.sdk.executor.get_required_model_classes_by_executor(
            aggr_id=AGGR_ID, version=VERSION
        )
        self.assertEqual(required_classes, {})

    def test_print_model_schema_for_executor(self):
        """Test a successful sdk.print_model_schema_for_executor call."""
        ML_management.sdk.executor.print_model_schema_for_executor(
            executor_aggr_id=AGGR_ID,
            models=[{"aggr_id": AGGR_ID, "version": VERSION, "role": ROLE}],
            executor_version=VERSION,
        )

    def test_print_datasetloader_schema(self):
        """Test a successful sdk.print_datasetloader_schema call."""
        ML_management.sdk.dataset_loader.print_dataset_loader_schema(name=NAME, version=VERSION)

    def test_print_executor_schema(self):
        """Test a successful sdk.print_executor_schema call."""
        ML_management.sdk.executor.print_executor_schema(aggr_id=AGGR_ID, version=VERSION)

    def test_print_executor_roles(self):
        """Test a successful sdk.print_executor_roles call."""
        ML_management.sdk.executor.print_executor_roles(aggr_id=AGGR_ID, version=VERSION)

    def test_cancels(self):
        """Test a successful sdk.cancel_* calls."""
        self.assertIsInstance(
            ML_management.sdk.model.cancel_build_job_for_model_version(aggr_id=AGGR_ID, model_version=VERSION), bool
        )
        self.assertIsInstance(
            ML_management.sdk.executor.cancel_build_job_for_executor_version(aggr_id=AGGR_ID, executor_version=VERSION),
            bool,
        )
        self.assertIsInstance(
            ML_management.sdk.model.cancel_venv_build_job_for_model_version(aggr_id=AGGR_ID, model_version=VERSION),
            bool,
        )
        self.assertIsInstance(ML_management.sdk.job.cancel_job(VERSION), bool)

    def test_delete_objects(self):
        """Test a successful sdk.delete_* calls."""
        self.assertIsInstance(
            ML_management.sdk.model.delete_model_version(model_name=NAME, model_version=VERSION), Union[ModelInfo, None]
        )
        self.assertIsInstance(ML_management.sdk.model.delete_model(model_name=NAME), bool)
        self.assertIsInstance(
            ML_management.sdk.executor.delete_executor_version(executor_name=NAME, executor_version=VERSION),
            Union[ExecutorInfo, None],
        )
        self.assertIsInstance(ML_management.sdk.executor.delete_executor(executor_name=NAME), bool)
        self.assertIsInstance(ML_management.sdk.dataset_loader.delete_dataset_loader(dataset_loader_name=NAME), bool)
        self.assertIsInstance(
            ML_management.sdk.dataset_loader.delete_dataset_loader_version(
                dataset_loader_name=NAME, dataset_loader_version=VERSION
            ),
            Union[DatasetLoaderInfo, None],
        )


if __name__ == "__main__":
    unittest.main()



"""Class inherits from base class of Model with a specific method for training the model."""
from abc import ABC, abstractmethod

from ML_management.model.patterns.model_pattern import Model


class TrainableModel(Model, ABC):
    """Implementation of trainable model."""

    def __init__(self):
        super().__init__()

    @abstractmethod
    def train_function(self, **kwargs):
        """Define train_function."""
        raise NotImplementedError
